# Copyright 2021 MosaicML. All Rights Reserved.

# type: ignore

import logging
from collections import Counter
from dataclasses import asdict, dataclass
from typing import Optional

import yahp as hp
from torch.optim.lr_scheduler import CosineAnnealingLR, CosineAnnealingWarmRestarts, ExponentialLR, MultiStepLR, StepLR

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State
from composer.core.types import Scheduler
from composer.optim.scheduler import ConstantLR
from composer.utils import ensure_tuple

log = logging.getLogger(__name__)


def scale_scheduler(scheduler: Scheduler, ssr: float, orig_max_epochs: Optional[int] = None):
    if isinstance(scheduler, StepLR):
        scheduler.step_size = int(scheduler.step_size * ssr)
    elif isinstance(scheduler, MultiStepLR):
        scheduler.milestones = Counter([int(ms * ssr) for ms in scheduler.milestones])
    elif isinstance(scheduler, CosineAnnealingLR):
        assert orig_max_epochs is not None, "To scale Cosine decay, max_epochs must be provided."

        # TODO: get warmup directly and impute unmodified T_max
        if hasattr(scheduler, 'interval') and scheduler.interval == "step":
            orig_max_epochs *= scheduler.steps_per_epoch

        warmup = orig_max_epochs - scheduler.T_max
        scheduler.T_max = int(orig_max_epochs * ssr - warmup)
    elif isinstance(scheduler, CosineAnnealingWarmRestarts):
        scheduler.T_0 = int(scheduler.T_0 * ssr)  # TODO: account for warmups
    elif isinstance(scheduler, ExponentialLR):
        factor = 1 / ssr
        scheduler.gamma = scheduler.gamma**factor
    elif isinstance(scheduler, ConstantLR):
        return
    elif hasattr(scheduler, 'scale_schedule') and callable(scheduler.scale_schedule):
        scheduler.scale_schedule(ssr)
    else:
        raise ValueError(f'Scale schedule being applied to unrecognized Scheduler {scheduler}. '
                         'Please implement scale_schedule(ssr: float) method in your scheduler.')


@dataclass
class ScaleScheduleHparams(AlgorithmHparams):
    ratio: float = hp.required('Ratio to scale the schedule.', template_default=1.0)
    method: str = hp.optional("Method to scale the schedule, one of 'epoch' or 'samples'. Default: epoch.",
                              default='epoch')

    def __post_init__(self):
        assert self.method in ('epoch', 'samples'), "Scale schedule method must be one of epoch or samples."

    def initialize_object(self) -> "ScaleSchedule":
        return ScaleSchedule(**asdict(self))


class ScaleSchedule(Algorithm):
    """ Scale the learning rate schedule

    Args:
        ratio (float): Ratio of full training schedule

        method (:obj: `str`, optional): Step or epoch, defaults to epoch
    """

    def __init__(self, ratio: float, method: str = 'epoch'):
        self.hparams = ScaleScheduleHparams(ratio=ratio, method=method)
        self.activated = False

    def match(self, event: Event, state: State) -> bool:
        return event == Event.TRAINING_START

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        # TODO: move the run once check as a decorator
        assert self.activated is False, "Scale Schedule should only be run once, check your control flow."
        assert state.schedulers is not None

        orig_max_epochs = state.max_epochs
        new_max_epochs = int(state.max_epochs * self.hparams.ratio)
        log.info(f'max_epochs changed from {state.max_epochs} to {new_max_epochs}')
        state.max_epochs = new_max_epochs
        if state.max_epochs == 0:
            raise ValueError('Scale schedule has reduced the max_epochs to 0. Set a higher ratio or more epochs.')

        if hasattr(state.schedulers, 'schedulers'):
            schedulers = state.schedulers.schedulers
        else:
            schedulers = ensure_tuple(state.schedulers)

        if self.hparams.method == 'epoch':
            for scheduler in schedulers:
                scale_scheduler(scheduler, self.hparams.ratio, orig_max_epochs)
        elif self.hparams.method == 'samples':
            raise NotImplementedError('Scale schedule algorithm with samples method not supported yet.')

        self.activated = True



# Copyright 2021 MosaicML. All Rights Reserved.

# Modified from https://github.com/google-research/augmix/blob/master/augmentations.py

"""Augmentation primitives for use in AugMix and RandAugment. Augmentation intensities
    are normalized on a scale of 1-10, where 10 is the strongest and maximum value an
    augmentation function will accept.
"""

import numpy as np
from PIL import Image, ImageEnhance, ImageOps


def int_parameter(level, maxval):
    """Helper function to scale a value between 0 and maxval and return as an int.
  Args:
    level: Level of the operation that will be between [0, `PARAMETER_MAX`].
    maxval: Maximum value that the operation can have. This will be scaled to
      level/PARAMETER_MAX.
  Returns:
    An int that results from scaling `maxval` according to `level`.
  """
    return int(level * maxval / 10)


def float_parameter(level, maxval):
    """Helper function to scale a value between 0 and maxval and return as a float.
  Args:
    level: Level of the operation that will be between [0, `PARAMETER_MAX`].
    maxval: Maximum value that the operation can have. This will be scaled to
        level/PARAMETER_MAX.
  Returns:
    A float that results from scaling `maxval` according to `level`.
  """
    return float(level) * maxval / 10.


def sample_level(n):
    """Helper function to sample from a uniform distribution between 0.1 and some value n."""
    return np.random.uniform(low=0.1, high=n)


def symmetric_sample(level):
    """Helper function to sample from a distribution over the domain [0.1, 10] with
        median == 1 and uniform probability of x | 0.1 ≤ x ≤ 1, and x | 1 ≤ x ≤ 10. Used
        for sampling transforms that can range from intensity 0 to infinity, and for
        which an intensity of 1 == no change.
    """
    if np.random.uniform() > 0.5:
        return np.random.uniform(1, level)
    else:
        return np.random.uniform(1 - (0.09 * level), 1)


def autocontrast(pil_img, _):
    return ImageOps.autocontrast(pil_img)


def equalize(pil_img, _):
    return ImageOps.equalize(pil_img)


def posterize(pil_img, level):
    level = int_parameter(sample_level(level), 4)
    return ImageOps.posterize(pil_img, 4 - level)


def rotate(pil_img, level):
    degrees = int_parameter(sample_level(level), 30)
    if np.random.uniform() > 0.5:
        degrees = -degrees
    return pil_img.rotate(degrees, resample=Image.BILINEAR)


def solarize(pil_img, level):
    level = int_parameter(sample_level(level), 256)
    return ImageOps.solarize(pil_img, 256 - level)


def shear_x(pil_img, level):
    level = float_parameter(sample_level(level), 0.3)
    if np.random.uniform() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, level, 0, 0, 1, 0), resample=Image.BILINEAR)


def shear_y(pil_img, level):
    level = float_parameter(sample_level(level), 0.3)
    if np.random.uniform() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, 0, 0, level, 1, 0), resample=Image.BILINEAR)


def translate_x(pil_img, level):
    level = int_parameter(sample_level(level), pil_img.size[0] / 3)
    if np.random.random() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, 0, level, 0, 1, 0), resample=Image.BILINEAR)


def translate_y(pil_img, level):
    level = int_parameter(sample_level(level), pil_img.size[1] / 3)
    if np.random.random() > 0.5:
        level = -level
    return pil_img.transform(pil_img.size, Image.AFFINE, (1, 0, 0, 0, 1, level), resample=Image.BILINEAR)


# The following augmentations overlap with corruptions in the ImageNet-C/CIFAR10-C test
# sets. Their original implementations also have an intensity sampling scheme that
# samples a value bounded by 0.118 at a minimum, and a maximum value of intensity*0.18+
# 1, which ranged from 0.28 (intensity = 1) to 1.9 (intensity 10). These augmentations
# have different effects depending on whether they are < 0 or > 0, so the original
# sampling scheme does not make sense to me. Accordingly, I replaced it with the
# symmetric_sample() above.


def color(pil_img, level):
    level = symmetric_sample(level)
    return ImageEnhance.Color(pil_img).enhance(level)


def color_original(pil_img, level):
    level = float_parameter(sample_level(level), 1.8) + 0.1
    return ImageEnhance.Color(pil_img).enhance(level)


def contrast(pil_img, level):
    level = symmetric_sample(level)
    return ImageEnhance.Contrast(pil_img).enhance(level)


def contrast_original(pil_img, level):
    level = float_parameter(sample_level(level), 1.8) + 0.1
    return ImageEnhance.Contrast(pil_img).enhance(level)


def brightness(pil_img, level):
    level = symmetric_sample(level)
    # Reduce intensity of brightness increases
    if level > 1:
        level = level * .75
    return ImageEnhance.Brightness(pil_img).enhance(level)


def brightness_original(pil_img, level):
    level = float_parameter(sample_level(level), 1.8) + 0.1
    return ImageEnhance.Brightness(pil_img).enhance(level)


def sharpness(pil_img, level):
    level = symmetric_sample(level)
    return ImageEnhance.Sharpness(pil_img).enhance(level)


def sharpness_original(pil_img, level):
    level = float_parameter(sample_level(level), 1.8) + 0.1
    return ImageEnhance.Sharpness(pil_img).enhance(level)


augmentation_sets = {
    "all": [
        autocontrast, equalize, posterize, rotate, solarize, shear_x, shear_y, translate_x, translate_y, color,
        contrast, brightness, sharpness
    ],
    # Augmentations that don't overlap with ImageNet-C/CIFAR10-C test sets
    "safe": [autocontrast, equalize, posterize, rotate, solarize, shear_x, shear_y, translate_x, translate_y],
    # Augmentations that use original implementations of color, contrast, brightness, and sharpness
    "original": [
        autocontrast, equalize, posterize, rotate, solarize, shear_x, shear_y, translate_x, translate_y, color_original,
        contrast_original, brightness_original, sharpness_original
    ],
}



# Copyright 2021 MosaicML. All Rights Reserved.

from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import List, Type

import torch
import torch_optimizer
import yahp as hp

from composer.core.types import ModelParameters, Optimizer
from composer.optim import DecoupledAdamW, DecoupledSGDW

# Optimizer parameters and defaults match those in torch.optim


@dataclass
class OptimizerHparams(hp.Hparams, ABC):
    """Abstract base class for optimizer hyperparameter classes.
    """

    @property
    @abstractmethod
    def optimizer_object(cls) -> Type[Optimizer]:
        pass

    def initialize_object(self, param_group: ModelParameters) -> Optimizer:
        assert issubclass(self.optimizer_object, torch.optim.Optimizer)
        return self.optimizer_object(param_group, **asdict(self))


@dataclass
class AdamHparams(OptimizerHparams):
    """Hyperparameters for the `Adam <https://pytorch.org/docs/stable/generated/torch.optim.Adam.html#torch.optim.Adam>`_
    optimizer.
    """
    lr: float = hp.optional(default=0.001, doc='learning rate')
    betas: List[float] = hp.optional(default_factory=lambda: [0.9, 0.999],
                                     doc='coefficients used for computing running averages of gradient and its square.')
    eps: float = hp.optional(default=1e-8, doc='term for numerical stability')
    weight_decay: float = hp.optional(default=0.0, doc='weight decay (L2 penalty)')
    amsgrad: bool = hp.optional(default=False, doc='use AMSGrad variant')

    @property
    def optimizer_object(cls) -> Type[torch.optim.Adam]:
        return torch.optim.Adam


@dataclass
class RAdamHparams(OptimizerHparams):
    """Hyperparameters for the `RAdam <https://pytorch.org/docs/1.10./generated/torch.optim.RAdam.html#torch.optim.RAdam>`_
    optimizer.
    """
    lr: float = hp.optional(default=0.001, doc='learning rate')
    betas: List[float] = hp.optional(default_factory=lambda: [0.9, 0.999],
                                     doc='coefficients used for computing running averages of gradient and its square.')
    eps: float = hp.optional(default=1e-8, doc='term for numerical stability')
    weight_decay: float = hp.optional(default=0.0, doc='weight decay (L2 penalty)')

    @property
    def optimizer_object(cls) -> Type[torch_optimizer.RAdam]:
        return torch_optimizer.RAdam


@dataclass
class AdamWHparams(OptimizerHparams):
    """ Hyperparameters for the `AdamW <https://pytorch.org/docs/stable/generated/torch.optim.AdamW.html#torch.optim.AdamW>`_
    optimizer.
    """
    lr: float = hp.optional(default=0.001, doc='learning rate')
    betas: List[float] = hp.optional(default_factory=lambda: [0.9, 0.999],
                                     doc='coefficients used for computing running averages of gradient and its square.')
    eps: float = hp.optional(default=1e-8, doc='term for numerical stability')
    weight_decay: float = hp.optional(default=1e-2, doc='weight decay (L2 penalty)')
    amsgrad: bool = hp.optional(default=False, doc='use AMSGrad variant')

    @property
    def optimizer_object(cls) -> Type[torch.optim.AdamW]:
        return torch.optim.AdamW


@dataclass
class DecoupledAdamWHparams(OptimizerHparams):
    """ Hyperparameters for the :class:`~composer.optim.DecoupledAdamW` optimizer.
    """
    lr: float = hp.optional(default=0.001, doc='learning rate')
    betas: List[float] = hp.optional(default_factory=lambda: [0.9, 0.999],
                                     doc='coefficients used for computing running averages of gradient and its square.')
    eps: float = hp.optional(default=1e-8, doc='term for numerical stability')
    weight_decay: float = hp.optional(default=1e-2, doc='weight decay (L2 penalty)')
    amsgrad: bool = hp.optional(default=False, doc='use AMSGrad variant')

    @property
    def optimizer_object(cls) -> Type[DecoupledAdamW]:
        return DecoupledAdamW


@dataclass
class SGDHparams(OptimizerHparams):
    """ Hyperparameters for the `SGD <https://pytorch.org/docs/stable/generated/torch.optim.SGD.html#torch.optim.SGD>`_
    optimizer.
    """
    lr: float = hp.required(doc='learning rate')
    momentum: float = hp.optional(default=0.0, doc='momentum factor')
    weight_decay: float = hp.optional(default=0.0, doc='weight decay (L2 penalty)')
    dampening: float = hp.optional(default=0.0, doc='dampening for momentum')
    nesterov: bool = hp.optional(default=False, doc='Nesterov momentum')

    @property
    def optimizer_object(cls) -> Type[torch.optim.SGD]:
        return torch.optim.SGD


@dataclass
class DecoupledSGDWHparams(OptimizerHparams):
    """ Hyperparameters for the :class:`~composer.optim.DecoupledSGDW` optimizer.
    """
    lr: float = hp.required(doc='learning rate')
    momentum: float = hp.optional(default=0.0, doc='momentum factor')
    weight_decay: float = hp.optional(default=0.0, doc='weight decay (L2 penalty)')
    dampening: float = hp.optional(default=0.0, doc='dampening for momentum')
    nesterov: bool = hp.optional(default=False, doc='Nesterov momentum')

    @property
    def optimizer_object(cls) -> Type[DecoupledSGDW]:
        return DecoupledSGDW


@dataclass
class RMSPropHparams(OptimizerHparams):
    """ Hyperparameters for the [RMSProp optimizer](https://pytorch.org/docs/stable/generated/torch.optim.RMSprop.html#torch.optim.RMSprop).
    """
    lr: float = hp.required(doc='learning rate')
    alpha: float = hp.optional(default=0.99, doc='smoothing constant')
    eps: float = hp.optional(default=1e-8, doc='term for numerical stability')
    momentum: float = hp.optional(default=0.0, doc='momentum factor')
    weight_decay: float = hp.optional(default=0.0, doc='weight decay (L2 penalty)')
    centered: bool = hp.optional(
        default=False,
        doc='normalize gradient by an estimation of variance',
    )

    @property
    def optimizer_object(cls) -> Type[torch.optim.RMSprop]:
        return torch.optim.RMSprop


def get_optimizer(param_groups: ModelParameters, hparams: OptimizerHparams) -> Optimizer:
    """ Get the optimizer specified by the given hyperparameters.

    Args:
        param_groups (ModelParameters): List of model parameters to optimize.
        hparams (OptimizerHparams): Instance of an optimizer's hyperparameters.
    """

    return hparams.initialize_object(param_group=param_groups)



# Copyright 2021 MosaicML. All Rights Reserved.

import dataclasses

import pytest

from composer.algorithms import (AlgorithmHparams, AlibiHparams, AugMixHparams, BlurPoolHparams, ChannelsLastHparams,
                                 ColOutHparams, CutOutHparams, DummyHparams, GhostBatchNormHparams,
                                 LabelSmoothingHparams, LayerFreezingHparams, MixUpHparams, NoOpModelHparams,
                                 ProgressiveResizingHparams, RandAugmentHparams, SAMHparams, ScaleScheduleHparams,
                                 SelectiveBackpropHparams, SeqLengthWarmupHparams, SqueezeExciteHparams,
                                 StochasticDepthHparams, SWAHparams, algorithm_registry)
from composer.core.algorithm import Algorithm

default_required_fields = {
    AlibiHparams: {
        'position_embedding_attribute': 'module.transformer.wpe',
        'attention_module_name': 'transformers.models.gpt2.modeling_gpt2.GPT2Attention',
        'attr_to_replace': '_attn',
        'alibi_attention': 'composer.algorithms.alibi.gpt2_alibi._attn',
        'mask_replacement_function': 'composer.algorithms.alibi.gpt2_alibi.enlarge_mask',
    },
    BlurPoolHparams: {
        'replace_convs': True,
        'replace_maxpools': True,
        'blur_first': True
    },
    LabelSmoothingHparams: {
        'alpha': 0.1
    },
    LayerFreezingHparams: {
        'freeze_start': 0.5,
        'freeze_level': 1.0
    },
    ChannelsLastHparams: {},
    ColOutHparams: {
        "p_row": 0.15,
        "p_col": 0.15,
        "batch": True,
    },
    SeqLengthWarmupHparams: {
        "duration": 0.30,
        "min_seq_length": 8,
        "max_seq_length": 1024,
        "step_size": 8,
        "truncate": True,
    },
    CutOutHparams: {
        'n_holes': 1,
        'length': 112
    },
    MixUpHparams: {
        'alpha': 0.2
    },
    DummyHparams: {},
    GhostBatchNormHparams: {
        'ghost_batch_size': 32
    },
    ScaleScheduleHparams: {
        'ratio': 0.5
    },
    NoOpModelHparams: {},
    SqueezeExciteHparams: {
        'latent_channels': 64,
        'min_channels': 128
    },
    StochasticDepthHparams: {
        'stochastic_method': 'block',
        'target_layer_name': 'ResNetBottleneck',
        'drop_rate': 0.2,
        'drop_distribution': 'linear',
        'drop_warmup': 0.0,
        'use_same_gpu_seed': False
    },
    ProgressiveResizingHparams: {
        'mode': 'resize',
        'initial_scale': 0.5,
        'finetune_fraction': 0.2,
        'resize_targets': False,
    },
    RandAugmentHparams: {
        'severity': 9,
        'depth': 2,
        'augmentation_set': 'all'
    },
    SWAHparams: {
        'swa_start': 0.8,
        'anneal_epochs': 10,
        'swa_lr': None
    },
    AugMixHparams: {
        'severity': 3,
        'depth': 3,
        'width': 3,
        'alpha': 1.0,
        'augmentation_set': 'all'
    },
    SAMHparams: {},
    SelectiveBackpropHparams: {
        'start': 0.5,
        'end': 0.9,
        'keep': 0.5,
        'scale_factor': 0.5,
        'interrupt': 2
    },
}


@pytest.fixture
def registry():
    return algorithm_registry.get_algorithm_registry()


@pytest.mark.parametrize("name", algorithm_registry.list_algorithms())
def test_algorithm_registry(name, registry):
    # create the hparams object
    hparams_class = registry[name]
    hparams = hparams_class(**default_required_fields[hparams_class])
    assert isinstance(hparams, AlgorithmHparams)
    assert dataclasses.is_dataclass(hparams)

    algorithm = algorithm_registry.get_algorithm(hparams)
    assert isinstance(algorithm, Algorithm)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.resnet18.model import ResNet18 as ResNet18
from composer.models.resnet18.resnet18_hparams import ResNet18Hparams as ResNet18Hparams

_task = 'Image Classification'
_dataset = 'ImageNet'
_name = 'ResNet18'
_quality = ''
_metric = 'Top-1 Accuracy'
_ttt = '?'
_hparams = 'hparams.yaml'



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest

from composer.core.state import State
from composer.loggers import Logger


@pytest.fixture()
def noop_dummy_logger(dummy_state: State) -> Logger:
    return Logger(state=dummy_state, backends=[])



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import asdict, dataclass

from composer.models.model_hparams import ModelHparams


@dataclass
class CIFARResNetHparams(ModelHparams):

    def initialize_object(self):
        from composer.models import CIFAR10_ResNet56
        return CIFAR10_ResNet56(**asdict(self))



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.sam.sam import SAM as SAM
from composer.algorithms.sam.sam import SAMHparams as SAMHparams

_name = 'SAM'
_class_name = 'SAM'
_functional = ''
_tldr = 'SAM optimizer measures sharpness of optimization space'
_attribution = '(Foret et al, 2021)'
_link = 'https://arxiv.org/abs/2010.01412'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.trainer.devices.device import Device as Device
from composer.trainer.devices.device_cpu import DeviceCPU as DeviceCPU
from composer.trainer.devices.device_gpu import DeviceGPU as DeviceGPU
from composer.trainer.devices.device_hparams import CPUDeviceHparams as CPUDeviceHparams
from composer.trainer.devices.device_hparams import DeviceHparams as DeviceHparams
from composer.trainer.devices.device_hparams import GPUDeviceHparams as GPUDeviceHparams



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import collections.abc
import contextlib
import logging
import warnings
from typing import Any, Dict, List, Optional, Sequence, Union

import torch
import torch.utils.data
from torch.backends import cudnn
from torch.cuda.amp.grad_scaler import GradScaler
from torch.nn.parallel import DistributedDataParallel
from torchmetrics.collections import MetricCollection
from torchmetrics.metric import Metric

from composer.core import Callback, Engine, Event, Logger, State
from composer.core.algorithm import Algorithm
from composer.core.logging import BaseLoggerBackend, LogLevel
from composer.core.types import Batch, BreakEpochException, Metrics, Precision, Tensor
from composer.datasets import DataloaderHparams, DataloaderSpec
from composer.loggers.tqdm_logger import TQDMLoggerBackend
from composer.models.base import BaseMosaicModel
from composer.optim import (ComposedScheduler, CosineAnnealingLRHparams, DecoupledSGDWHparams, OptimizerHparams,
                            SchedulerHparams, WarmUpLRHparams)
from composer.optim.scheduler import ensure_warmup_last
from composer.trainer.checkpoint import Checkpointer, CheckpointLoader
from composer.trainer.ddp import DDP, DataloaderMultipleIterationWarning, StoreHparams, TCPStoreHparams
from composer.trainer.devices.device import Device
from composer.trainer.devices.device_cpu import DeviceCPU
from composer.trainer.scaler import ClosureGradScaler
from composer.trainer.trainer_hparams import TrainerHparams
from composer.utils import ensure_tuple, get_random_seed, map_collection, seed_all

log = logging.getLogger(__name__)


class Trainer:
    """Trainer for training a model with algorithms.

    Can be created either with ``__init__`` or by providing a
    :class:`~composer.trainer.TrainerHparams` object
    (see :meth:`~composer.trainer.Trainer.create_from_hparams`).

    Args:
        model (BasicMosaicModel): The model to train.
        train_dataloader_spec (DataloaderSpec): The dataloader spec for the training data.
        eval_dataloader_spec (DataloaderSpec): The dataloader spec for the evaluation data.
        max_epochs (int): The maxmimum number of epochs to train for.
        train_batch_size (int): Minibatch size for training data.
        eval_batc_size (int): Minibatch size for evaluation data.
        algorithms (List[Algorithm], optional): The algorithms to use during training.
            (default: ``[]``)
        optimizer_hparams: (OptimizerHparams, optional): The OptimizerHparams for constructing
            the optimizer for training. Must pass OptimizerHparams instead of a `torch.optim.Optimizer`
            object because the optimizer has to be constructed after certain algorithms which modify
            the model architecture have run on the model. (default:
            ``MosaicMLSGDWHparams(lr=0.1, momentum=0.9, weight_decay=1.0e-4)``)
        schedulers_hparams: (Union[SchedulerHparams, List[SchedulerHparams]], optional): The
            SchedulerHparams for constructing the one or more learning rate schedulers used
            during training. Must pass SchedulerHparams instead of a `torch.optim.lr_scheduler._LRScheduler`
            object because the scheduler needs an optimizer to be constructed and we construct the optimizer
            in `__init__`. (default:
            ``[CosineAnnealingLRHparams(T_max=f"{max_epochs}ep"), WarmUpLRHparams()]``).
        device (Device, optional): The device to use for training. Either `DeviceCPU` or `DeviceGPU`.
            (default ``DeviceCPU(n_cpus=1)``)
        grad_accum (int, optional): The number of microbatches to split a per-device batch into. Gradients
            are summed over the microbatches per device. (default: ``1``)
        grad_clip_norm (float, optional): The norm to clip gradient magnitudes to. Set to None for no gradient
            clipping. (default: ``None``)
        validate_every_n_batches (int, optional): Compute metrics on evaluation data every N batches.
             Set to -1 to never validate on a batchwise frequency. (default: ``-1``)
        validate_every_n_epochs (int, optional): Compute metrics on evaluation data every N epochs.
            Set to -1 to never validate on a epochwise frequency. (default: ``1``)
        compute_training_metrics (bool, optional): True to compute metrics on training data and False to not.
            (default: ``False``)
        precision (Precision, optional): Numerical precision to use for training. (default: ``Precision.FP32``).
        num_workers (int, optional): The number of CPU workers to use per GPU. 0 results in loading data
            on the main process. (default: ``0``)
        prefetch_factor (int, optional): Number of samples loaded in advance by each worker. (default: ``2``)
        persistent_workers (bool, optional): Whether or not to shutdown workers after the dataset
            has been consumed once. (default: ``False``)
        pin_memory (bool, optional): Whether or not to copy data tensors into CUDA pinned memory.
            (default: ``False``)
        timeout (int, optional): Timeout value for collecting a batch from workers. 0 for no timeout.
            (default: ``0``)
        ddp_store_hparams (StoreHparams, optional): `DistributedDataParallel` configuration.
            (default: ``TCPStoreHparams("127.0.0.1", 43297)``)
        fork_rank_0 (bool, optional): True to fork the rank 0 process in distributed data parallel,
            False to not. (default: ``True``)
        seed (int, optional): The seed used in randomization. When not provided a random seed
            will be created. (default: ``None``)
        deterministic_mode (bool, optional): Run the model deterministically. Experimental. Performance
            degradations expected. Certain Torch modules may not have deterministic implementations,
            which will result in a crash. (default: ``False``)
        log_destinations (List[BaseLoggerBackend], optional): The destinations to log training information to.
            (default ``[TQDMLoggerBackend()]``).
        callbacks (Sequence[Callback], optional): The callbacks to run during training. (default: ``[]``)
        checkpoint_filepath (str, optional): The path to a trainer checkpoint file. If provided
            the trainer will load the state (along with it's associated attributes) during initialization.
            (default: ``None``)
        checkpoint_interval_unit (int, optional): Unit for the checkpoint save interval -- should be 'ep'
            for epochs, 'ba' for batches, or None to disable checkpointing. (default: ``None``).
        checkpoint_folder (str, optional): The folder to save checkpoints to. (default: ``checkpoints``)
        checkpoint_interval (int, optional): The frequency with which to checkpoint. (default: ``1``)
        config (Dict[str, Any], optional): Extra user-provided trainer configuration. Will be persisted
            along with the trainer state during checkpointing. (default: ``None``)

    Attributes:
        state (State): The :class:`State` object used to store training state.
        logger (Logger): The :class:`Logger` used for logging.
        engine (Engine): The :class:`Engine` used for running callbacks and algorithms.
    """

    def __init__(
            self,
            *,
            model: BaseMosaicModel,
            train_dataloader_spec: DataloaderSpec,
            eval_dataloader_spec: DataloaderSpec,
            max_epochs: int,
            train_batch_size: int,
            eval_batch_size: int,
            algorithms: Optional[List[Algorithm]] = None,
            optimizer_hparams: Optional[OptimizerHparams] = None,
            schedulers_hparams: Optional[Union[SchedulerHparams, List[SchedulerHparams]]] = None,

            # device
            device: Optional[Device] = None,

            # training hparams
            grad_accum: int = 1,
            grad_clip_norm: Optional[float] = None,
            validate_every_n_batches: int = -1,
            validate_every_n_epochs: int = 1,
            compute_training_metrics: bool = False,
            precision: Precision = Precision.FP32,

            # dataloader hparams
            num_workers: int = 0,
            prefetch_factor: int = 2,
            persistent_workers: bool = False,
            pin_memory: bool = False,
            timeout: int = 0,

            # ddp hparams
            ddp_store_hparams: Optional[StoreHparams] = None,
            fork_rank_0: bool = True,

            # Randomness
            seed: Optional[int] = None,
            deterministic_mode: bool = False,

            # Logging and callbacks
            log_destinations: Optional[List[BaseLoggerBackend]] = None,
            callbacks: Sequence[Callback] = tuple(),

            # Checkpoint hparams
            checkpoint_filepath: Optional[str] = None,
            checkpoint_interval_unit: Optional[str] = None,
            checkpoint_folder: Optional[str] = "checkpoints",
            checkpoint_interval: Optional[int] = 1,

            # Optional config (ex. an hparams yaml file)
            config: Optional[Dict[str, Any]] = None):
        # surpressing GradScaler warnings as they are always created
        # self._use_grad_scaling() will raise a RuntimeError if grad scaling is not available when it is required
        warnings.filterwarnings(action="ignore", message="torch.cuda.amp.GradScaler")

        self.config = config

        if not device:
            device = DeviceCPU(num_cpus=1)
        self.device = device

        if not seed:
            # Set a deterministic seed in the hparams
            # This seed will be dumped in the hparams that are saved with checkpoints
            seed = get_random_seed()
            log.info(f"Seed was None. Setting seed to random value: {seed}")
        # If hparams is used to create the Trainer this function is called twice
        # which is okay because all runs with the hparams codepath will do this
        seed_all(seed)

        if not algorithms:
            algorithms = []

        find_unused_parameters = any(map(lambda x: x.find_unused_parameters, algorithms))
        if not ddp_store_hparams:
            ddp_store_hparams = TCPStoreHparams("127.0.0.1", 43297)
        self.ddp = DDP(
            nproc_per_node=self.device.nproc_per_node,
            store_hparams=ddp_store_hparams,
            node_rank=0,
            num_nodes=1,
            backend=self.device.ddp_backend,
            fork_rank_0=fork_rank_0,
            find_unused_parameters=find_unused_parameters,
        )

        self.state = State(max_epochs=max_epochs,
                           train_batch_size=train_batch_size,
                           eval_batch_size=eval_batch_size,
                           algorithms=algorithms,
                           callbacks=callbacks,
                           model=model,
                           grad_accum=grad_accum,
                           precision=precision,
                           precision_context=self.device.precision_context,
                           nproc_per_node=self.device.nproc_per_node,
                           world_size=self.ddp.world_size,
                           seed=seed)

        if not log_destinations:
            log_destinations = [TQDMLoggerBackend()]
        self.logger = Logger(self.state, log_destinations)
        self.state.callbacks = [*log_destinations, *callbacks]

        self.engine = Engine(self.state, self.state.algorithms, self.logger, self.state.callbacks)

        self.train_dl_spec = train_dataloader_spec
        self.eval_dl_spec = eval_dataloader_spec

        self.dl_hparams = DataloaderHparams(num_workers=num_workers,
                                            prefetch_factor=prefetch_factor,
                                            persistent_workers=persistent_workers,
                                            pin_memory=pin_memory,
                                            timeout=timeout)

        self.validate_every_n_batches = validate_every_n_batches
        self.validate_every_n_epochs = validate_every_n_epochs
        self.compute_training_metrics = compute_training_metrics
        self.grad_clip_norm = grad_clip_norm

        if deterministic_mode:
            torch.use_deterministic_algorithms(True)
            cudnn.benchmark = False
            warnings.warn("Deterministic mode is activated. This will negatively impact performance.",
                          category=UserWarning)

        # TODO: Resolve https://github.com/mosaicml/mosaicml/pull/169#discussion_r703683025
        # Possibly rename Event name
        # run INIT event before optimizers and schedulers are created
        self.engine.run_event(Event.INIT)

        assert isinstance(self.train_dl_spec.dataset, collections.abc.Sized)
        steps_per_epoch = len(self.train_dl_spec.dataset) // train_batch_size
        # Need to use hparams here because optimizer and schedulers need to be created after Event.INIT
        if not optimizer_hparams:
            optimizer_hparams = DecoupledSGDWHparams(lr=0.1, momentum=0.9, weight_decay=1.0e-4)
        if not schedulers_hparams:
            schedulers_hparams = [CosineAnnealingLRHparams(T_max=f"{max_epochs}ep"), WarmUpLRHparams()]
        if not isinstance(schedulers_hparams, list):
            schedulers_hparams = [schedulers_hparams]
        optimizer = optimizer_hparams.initialize_object(param_group=self.state.model.parameters())
        schedulers = [x.initialize_object(optimizer, steps_per_epoch) for x in ensure_warmup_last(schedulers_hparams)]
        self.state.optimizers = optimizer
        self.state.schedulers = ComposedScheduler(schedulers=schedulers)

        self.checkpointer = None
        if checkpoint_folder and checkpoint_interval and checkpoint_interval_unit:
            self.checkpointer = Checkpointer(checkpoint_folder=checkpoint_folder,
                                             checkpoint_interval=checkpoint_interval,
                                             checkpoint_interval_unit=checkpoint_interval_unit)

        self.checkpoint_loader = None
        if checkpoint_filepath:
            self.checkpoint_loader = CheckpointLoader(checkpoint_filepath=checkpoint_filepath)
            self.checkpoint_loader.load_checkpoint(state=self.state)

    @classmethod
    def create_from_hparams(cls, hparams: TrainerHparams) -> Trainer:
        """Instantiate a Trainer using a `TrainerHparams` object.

        Args:
            hparams (TrainerHparams): The TrainerHparams object used to instantiate the trainer.

        Returns:
            A Trainer object initialized with the provided TrainerHparams.
        """

        hparams.validate()

        # devices and systems
        device = hparams.device.initialize_object()

        seed = hparams.seed if hparams.seed else get_random_seed()
        # need to set seed before model initialization for determinism
        seed_all(seed)

        model = hparams.model.initialize_object()
        algorithms = [x.initialize_object() for x in hparams.algorithms]

        # callbacks, loggers, and seed
        callbacks = [x.initialize_object() for x in hparams.callbacks]
        dict_config = hparams.to_dict()
        log_destinations = [x.initialize_object(config=dict_config) for x in hparams.loggers]

        find_unused_parameters = any(map(lambda x: x.find_unused_parameters, algorithms))
        ddp = hparams.ddp.initialize_object(
            device.nproc_per_node,
            device.ddp_backend,
            find_unused_parameters,
        )

        train_dl_spec = hparams.train_dataset.initialize_object()
        eval_dl_spec = hparams.val_dataset.initialize_object()

        trainer = cls(
            model=model,
            train_dataloader_spec=train_dl_spec,
            eval_dataloader_spec=eval_dl_spec,
            max_epochs=hparams.max_epochs,
            train_batch_size=hparams.total_batch_size,
            eval_batch_size=hparams.eval_batch_size,
            algorithms=algorithms,
            optimizer_hparams=hparams.optimizer,
            schedulers_hparams=hparams.schedulers,

            # device
            device=device,

            # training hparams
            grad_accum=hparams.grad_accum,
            grad_clip_norm=hparams.grad_clip_norm,
            validate_every_n_batches=hparams.validate_every_n_batches,
            validate_every_n_epochs=hparams.validate_every_n_epochs,
            compute_training_metrics=hparams.compute_training_metrics,
            precision=hparams.precision,

            # dataloader hparams
            num_workers=hparams.dataloader.num_workers,
            prefetch_factor=hparams.dataloader.prefetch_factor,
            persistent_workers=hparams.dataloader.persistent_workers,
            pin_memory=hparams.dataloader.pin_memory,
            timeout=hparams.dataloader.timeout,

            # ddp hparams
            ddp_store_hparams=ddp.store_hparams,
            fork_rank_0=ddp.fork_rank_0,

            # Randomness
            seed=seed,
            deterministic_mode=hparams.deterministic_mode,

            # Callbacks and logging
            log_destinations=log_destinations,
            callbacks=tuple(callbacks),

            # Checkpointing hparams
            checkpoint_filepath=hparams.checkpoint_filepath,
            checkpoint_interval_unit=hparams.checkpoint_interval_unit,
            checkpoint_folder=hparams.checkpoint_folder,
            checkpoint_interval=hparams.checkpoint_interval,

            # Optional config
            config=hparams.to_dict())

        return trainer

    def fit(self):
        """Train and evaluate the model on the provided data.
        """
        self.ddp.launch(self.state, self._train_loop)

    def _create_dataloaders(self) -> None:
        """Create the dataloaders.

        Should be called after distributed training has started,
        since the dataloader samplers need to know their rank.
        """
        # shorthand
        state = self.state

        # compute per gpu batch size
        train_gpu_batch_size = state.train_batch_size // state.world_size
        eval_gpu_batch_size = state.eval_batch_size // state.world_size

        train_dataloader = self.ddp.create_dataloader(train_gpu_batch_size, self.dl_hparams, self.train_dl_spec)
        eval_dataloader = self.ddp.create_dataloader(eval_gpu_batch_size, self.dl_hparams, self.eval_dl_spec)

        # move to device
        state.train_dataloader = self.device.dataloader_to_device(
            train_dataloader,
            self.train_dl_spec.prefetch_fn,
        )
        state.eval_dataloader = self.device.dataloader_to_device(
            eval_dataloader,
            self.eval_dl_spec.prefetch_fn,
        )

    def _get_metrics_as_collection(self, *, is_train: bool) -> MetricCollection:
        """Get metrics relevant to the model. Metrics are all implmented as subclasses
        of :class:`torchmetrics.Metric`. This function returns metrics as a
        :class:`~torchmetrics.collections.MetricCollection` to enable support
        for multiple metrics.

        Args:
            is_train (bool): True to get training metrics and false to get
            evaluation metrics.

        Returns:
            A :class:`~torchmetrics.collections.MetricCollection` object.
        """
        original_model = self.state.model.module
        assert isinstance(original_model, BaseMosaicModel)

        metrics = original_model.metrics(train=is_train)
        assert isinstance(metrics, (Metric, MetricCollection)), \
            "Error module.metrics() must return a Metric or MetricCollection object."
        if isinstance(metrics, Metric):
            # Forcing metrics to be a MetricCollection simplifies logging results
            metrics = MetricCollection([metrics])

        # Safety check to ensure the metric and data are on the same device. Normally not
        # needed because the metric is automatically on the same device as the model.
        # See https://torchmetrics.readthedocs.io/en/latest/pages/overview.html for details.
        metrics = self.device.module_to_device(metrics)
        return metrics

    def _compute_and_log_metrics(self, metrics: Metrics, *, is_train: bool, is_batch: bool):
        """Computes metrics, logs the results, and resets the metrics.

        Args:
            metrics (Metrics): The metrics to compute.
            is_train (bool): True for training metrics, False for evaluation metrics.
            is_batch (bool): True if logging at batch level, false for epoch level.
        """
        computed_metrics = metrics.compute()
        for name, value in computed_metrics.items():
            log_level = LogLevel.BATCH if is_batch else LogLevel.EPOCH
            suffix = 'train' if is_train else 'val'
            self.logger.metric(log_level, {f'{name.lower()}/{suffix}': value})
        metrics.reset()

    def _spin_dataloaders(self):
        """Spin the dataloaders to restore sampler state.

        Only one batch must be loaded to seed the sampler's generator.
        since only the first batch is being loaded, the dataloader may
        not be completely iterated through.
        """
        # surpressing this multiple iteration warning -- it is OK to ignore
        warnings.simplefilter(action="ignore", category=DataloaderMultipleIterationWarning, append=True)
        assert self.state.train_dataloader is not None, "train dataloader should be set"
        assert self.state.eval_dataloader is not None, "eval dataloader should be set"

        # spin the eval dataloader once to initialize its sampler deterministically
        # so it does not affect any other RNG reads
        for _ in self.state.eval_dataloader:
            break

        # spin the train dataloader's sampler to get to the state of the desired epoch
        for _ in range(self.state.epoch):
            for _ in self.state.train_dataloader:
                break

    def _get_batch_size(self, batch: Batch) -> int:
        if isinstance(batch, Tensor):
            return batch.shape[0]

        dim0_sizes = []
        if isinstance(batch, (list, tuple)):
            dim0_sizes = [t.shape[0] for t in batch]
        elif isinstance(batch, dict):
            dim0_sizes = [t.shape[0] for t in batch.values()]

        if len(set(dim0_sizes)) == 1:
            return dim0_sizes[0]
        else:
            raise ValueError('The default _get_batch_size function found ',
                             f'multiple Tensor sizes in batch: {dim0_sizes}')

    def _train_loop(self) -> None:
        """Run training for the specified number of epochs and log results.
        """
        # shorthand
        state = self.state

        assert state.optimizers is not None
        assert state.schedulers is not None
        # place the state, model in the proper devices
        self.device.prepare(state)
        state.model = self.device.module_to_device(state.model)
        state.optimizers = map_collection(state.optimizers, self.device.optimizer_to_device)

        # create dataloaders here after distributed training has started
        self._create_dataloaders()
        if state.train_dataloader is None or state.eval_dataloader is None:
            raise ValueError('Dataloaders were not created properly, and are None.')

        # wrap model with DDP
        state.model = self.ddp.prepare_module(state.model)
        original_model = state.model.module
        assert isinstance(original_model, BaseMosaicModel)

        # print training start
        self.logger.metric_fit({"trainer/algorithms": [str(algo) for algo in self.engine.algorithms]})

        if self.compute_training_metrics:
            log.warn('Computing model evaluation metrics during training.'
                     ' This doubles the number of forward passes and may lead'
                     ' to a throughput degradation.')
        train_metrics = self._get_metrics_as_collection(is_train=True)

        self.engine.run_event(Event.TRAINING_START)

        if len(ensure_tuple(state.optimizers)) != 1:
            raise NotImplementedError("The Mosaic trainer only supports one optimizer; "
                                      f"found {len(ensure_tuple(state.optimizers))} optimizers")

        if self._use_closures():

            def _ddp_reduce_scalar_and(flag: bool) -> bool:
                value = 1 if flag else 0
                flag_tensor = self.device.tensor_to_device(torch.tensor(value).int())
                self.ddp.all_reduce(flag_tensor, reduce_operation='PRODUCT')
                return flag_tensor.item() == 1

            def _ddp_reduce_tensor_sum(tensor: Tensor) -> Tensor:
                # Happens in-place; that's fine
                self.ddp.all_reduce(tensor, reduce_operation="SUM")
                return tensor

            state.scaler = ClosureGradScaler(ddp_reduce_scalar_and=_ddp_reduce_scalar_and,
                                             ddp_reduce_tensor_sum=_ddp_reduce_tensor_sum)
        else:
            state.scaler = GradScaler()
        use_grad_scaling = self._use_grad_scaling(state.precision, state.scaler)

        self._spin_dataloaders()

        if self.state.batch_idx == 0 and self.checkpoint_loader:
            # only restore the rng state here if the step in the current epoch is zero.
            self.checkpoint_loader.restore_checkpoint_rng_state(self.state, self.device)

        for _ in range(state.epoch, state.max_epochs):
            try:
                state.model.train()

                if self.state.batch_idx == 0:
                    self.engine.run_event(Event.EPOCH_START)
                    self.logger.metric_epoch({"epoch": self.state.epoch})

                assert state.train_dataloader, "Train Dataloader must be set"
                for batch_idx, state.batch in enumerate(state.train_dataloader):

                    # if resuming, skip dataloader forward to the minibatch index
                    if batch_idx < self.state.batch_idx:
                        if self.checkpoint_loader:
                            self.checkpoint_loader.restore_checkpoint_rng_state(self.state, self.device)
                        continue

                    state.last_batch_size = self._get_batch_size(state.batch)

                    if self.compute_training_metrics:
                        # compute metrics on the training set
                        state.model.eval()
                        with torch.no_grad():
                            eval_microbatches = self.train_dl_spec.split_fn(state.batch, state.grad_accum)
                            for eval_microbatch in eval_microbatches:
                                # TODO: Detect if self.run_event(Event.AFTER_DATALOADER) changes the training
                                # data and if so print a warning that metrics may return unexpected results
                                outputs, targets = original_model.validate(eval_microbatch)
                                train_metrics.update(outputs, targets)

                    state.model.train()

                    self.engine.run_event(Event.AFTER_DATALOADER)

                    microbatches = self.train_dl_spec.split_fn(state.batch, state.grad_accum)

                    self.engine.run_event(Event.BATCH_START)
                    self.logger.metric_batch({
                        "trainer/global_step": self.state.step,
                        "trainer/batch_idx": self.state.batch_idx,
                    })
                    total_loss = None
                    if self._use_closures():
                        closure = lambda **kwargs: self._train_batch(microbatches, **kwargs)
                        for optimizer in ensure_tuple(state.optimizers):
                            if use_grad_scaling:
                                total_loss = state.scaler.step(optimizer, closure=closure)
                            else:
                                # Torch optimizers technically expect closures to return a float, not a Tensor.
                                # In practice, this doesn't seem to actually matter.
                                total_loss = optimizer.step(closure=closure)  # type: ignore
                    else:
                        total_loss = self._train_batch(microbatches)
                        for optimizer in ensure_tuple(state.optimizers):
                            if use_grad_scaling:
                                state.scaler.step(optimizer)
                            else:
                                optimizer.step()

                    if use_grad_scaling:
                        state.scaler.update()

                    if total_loss is not None:
                        assert isinstance(total_loss, Tensor)

                        # total_loss can be None if gradient scaling failed
                        self.ddp.all_reduce(total_loss, reduce_operation="SUM")
                        self.ddp.barrier()
                        full_loss = total_loss.cpu().item()
                        self.logger.metric_batch({'loss/train': full_loss / state.world_size})

                    if self.compute_training_metrics:
                        self._compute_and_log_metrics(train_metrics, is_train=True, is_batch=True)

                    self.engine.run_event(Event.BATCH_END)

                    state.schedulers.step(interval='batch')  # type: ignore # TODO fix later

                    if self.validate_every_n_batches > 0 and (state.step + 1) % self.validate_every_n_batches == 0:
                        self.eval(is_batch=True)

                    state.step += 1
                    if self.checkpointer and self.checkpointer.should_checkpoint(state=state, event=Event.BATCH_END):
                        self.checkpointer.save_checkpoint(state=state,
                                                          device=self.device,
                                                          ddp=self.ddp,
                                                          config=self.config)
            except BreakEpochException:
                log.info(f'Skipping the rest of Epoch {state.epoch}')

            state.schedulers.step(interval='epoch')  # type: ignore # TODO fix later
            self.engine.run_event(Event.EPOCH_END)

            if self.validate_every_n_epochs > 0 and (state.epoch + 1) % self.validate_every_n_epochs == 0:
                self.eval(is_batch=False)

            state.epoch += 1

            if self.checkpointer and self.checkpointer.should_checkpoint(state=state, event=Event.EPOCH_END):
                self.checkpointer.save_checkpoint(state=state, device=self.device, ddp=self.ddp, config=self.config)

        self.engine.run_event(Event.TRAINING_END)

    def _train_batch(self, microbatches: Sequence[Batch], ddp_sync: bool = True):
        """Run training on a full batch of data.

        Args:
            microbatches (Sequence[Batch]): The microbatches which make up the batch.
            ddp_sync (bool): True to sync gradients between devices on every backwards
                pass and False to only sync gradients after each device has finished
                computing a gradient on it's entire set of microbatches. (default: ``True``)
        """
        if ddp_sync or not isinstance(self.state.model, DistributedDataParallel):
            context = contextlib.nullcontext
        else:
            context = self.state.model.no_sync

        with context():  # type: ignore - Pyright apparently doesn't recognize no_sync
            return self._train_batch_inner(microbatches)

    def _train_batch_inner(self, microbatches: Sequence[Batch]):
        """Iterate over microbatches and compute the loss that will be used to step
        the optimizer.

        Args:
            microbatches (Sequence[Batch]): The microbatches which make up the batch.
        """
        self.engine.run_event(Event.BEFORE_TRAIN_BATCH)
        find_unused_parameters = any(map(lambda x: x.find_unused_parameters, self.engine.algorithms))

        state = self.state
        original_model = state.model.module
        assert isinstance(original_model, BaseMosaicModel)
        assert state.optimizers is not None
        assert state.scaler is not None

        use_grad_scaling = self._use_grad_scaling(state.precision, state.scaler)

        for optimizer in ensure_tuple(state.optimizers):
            optimizer.zero_grad()

        # tracker for gradient accumulation
        total_loss = self.device.tensor_to_device(torch.zeros(size=(1,)))
        current_batch_size = sum([self._get_batch_size(batch) for batch in microbatches])

        for microbatch_idx, state.batch in enumerate(microbatches):
            # It's only necessary to run a DDP sync on the final microbatch, since the synced
            # gradients aren't needed until after this function has finished.
            # When any algorithm sets find_unused_parameters to true, DDP must sync every microbatch.

            if microbatch_idx + 1 == len(microbatches) or not isinstance(
                    state.model, DistributedDataParallel) or find_unused_parameters:
                context = contextlib.nullcontext
            else:
                context = state.model.no_sync
            with context():  # type: ignore - Pyright does not recognize type of no_sync
                last_microbatch_size = self._get_batch_size(state.batch)

                # forward pass
                self.engine.run_event(Event.BEFORE_FORWARD)

                with state.precision_context(state.precision):
                    state.outputs = state.model.forward(state.batch)

                self.engine.run_event(Event.AFTER_FORWARD)

                # loss
                self.engine.run_event(Event.BEFORE_LOSS)

                with state.precision_context(state.precision):
                    state.loss = original_model.loss(state.outputs, state.batch)

                for loss in ensure_tuple(state.loss):
                    loss.mul_(last_microbatch_size / current_batch_size)

                # Loss is added to losses with clone to not scale the loss for the step printout
                # Likely need to look into the performance impact
                for loss in ensure_tuple(state.loss):
                    total_loss += loss.detach().clone()

                assert state.loss is not None
                self.engine.run_event(Event.AFTER_LOSS)

                # backward
                self.engine.run_event(Event.BEFORE_BACKWARD)

                if use_grad_scaling:
                    state.loss = state.scaler.scale(state.loss)

                for loss in ensure_tuple(state.loss):
                    loss.backward()

                self.engine.run_event(Event.AFTER_BACKWARD)

        # Unscale gradients before `Event.AFTER_TRAIN_BATCH`
        if use_grad_scaling:
            for optimizer in ensure_tuple(state.optimizers):
                state.scaler.unscale_(optimizer)

        # clip gradients if the magnitude is too large
        if self.grad_clip_norm is not None:
            torch.nn.utils.clip_grad_norm_(
                parameters=state.model.parameters(),
                max_norm=self.grad_clip_norm,
            )

        self.engine.run_event(Event.AFTER_TRAIN_BATCH)

        return total_loss

    def eval(self, is_batch: bool):
        """Evaluate the model on the provided evaluation data and log
        appropriate metrics.

        Args:
            is_batch (bool): True to log metrics with ``LogLevel.BATCH``
                and False to log metrics with ``LogLevel.EPOCH``.
        """
        state = self.state
        model = state.model

        restore_model_train = model.training

        model.eval()
        with torch.no_grad():

            self.engine.run_event(Event.EVAL_START)

            original_model = state.model.module
            assert isinstance(original_model, BaseMosaicModel)

            metrics = self._get_metrics_as_collection(is_train=False)

            assert state.eval_dataloader is not None

            for state.batch in state.eval_dataloader:
                self.engine.run_event(Event.EVAL_BATCH_START)

                self.engine.run_event(Event.EVAL_BEFORE_FORWARD)
                state.outputs, targets = original_model.validate(state.batch)
                self.engine.run_event(Event.EVAL_AFTER_FORWARD)

                metrics.update(state.outputs, targets)

                self.engine.run_event(Event.EVAL_BATCH_END)

            self._compute_and_log_metrics(metrics, is_train=False, is_batch=is_batch)
            self.engine.run_event(Event.EVAL_END)

        if restore_model_train:
            model.train()

    def _use_grad_scaling(self, precision: Union[str, Precision], scaler: Optional[GradScaler]) -> bool:
        """Determines based on precision when to use grad scaling.

        By default, the pytorch GradScaler is a no-op if running on
        unsupported hardware. Here we raise a RuntimeError instead.

        Args:
            precision (Precision): Numerical precision, based on the Precision Enum.
            scaler (GradScaler): Used to make sure that the scaler is enabled when
            using grad scaling.

        Raises:
            RuntimeError:
                Occurs when attemptng to use grad scaling without the scaler
                enabled. Likely due to hardware not supporting the provided precision.
        """
        precision = Precision(precision)
        use_grad_scaling = precision == Precision.AMP

        if use_grad_scaling and (scaler is None or not scaler.is_enabled()):
            raise RuntimeError(f'Attempting to use grad scaling with {precision}, but scaler is not enabled.'
                               f'Potentially your hardware does not support Precision {precision}.')
        return use_grad_scaling

    def _use_closures(self) -> bool:
        """Determines based on precision and optimizers whether to use closures.

        We default to using closures unless AMP is enabled, in which case we only allow
        closures when using optimizers with the _step_supports_amp_closure flag.
        """
        if self.state.precision != Precision.AMP:
            return True

        if self.state.optimizers is None:
            raise RuntimeError("state.optimizers must be set before `_use_closures` can be determined")

        return all(
            getattr(optimizer, "_step_supports_amp_closure", False)
            for optimizer in ensure_tuple(self.state.optimizers))



from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from typing import List

import yahp as hp

from composer import Logger, State
from composer.callbacks.callback_hparams import CallbackHparams
from composer.core.callback import Callback
from composer.core.types import BreakEpochException

log = logging.getLogger(__name__)


class TimingMonitor(Callback):

    def __init__(self, min_steps: int, epoch_list: List[int], step_list: List[int], all_epochs: bool):
        super().__init__()
        self.hparams = TimingMonitorHparams(min_steps=min_steps,
                                            epoch_list=epoch_list,
                                            step_list=step_list,
                                            all_epochs=all_epochs)

        if not all_epochs:
            assert len(epoch_list) > 0, "'epoch_list'  must be non-empty."
            assert 0 in epoch_list, \
                "'epoch_list' must contain 0, otherwise first K epochs have unknown speed"
        assert len(step_list) > 0, "'step_list'  must be non-empty."
        assert 0 in step_list, \
            "'step_list' must contain 0 because `EPOCH_START` requires batch_idx of 0"

        # Sort lists so that time only moves forward
        epoch_list = list(sorted(epoch_list))
        step_list = list(sorted(step_list))

        self.current_time = None
        self.profile_examples = 0
        self.profile_steps = 0
        self.profile_time = 0
        self.wall_clock_train = 0

        self.min_steps = min_steps

        self.all_epochs = all_epochs
        self.epoch_list = epoch_list
        self.epoch_ix = 0
        self.step_list = step_list
        self.step_ix = 0

        # initialized at training_start
        self.original_max_epochs = -1
        self.wct_dict = {}

    def _compute_elapsed_wct(self, epoch_wct_dict, steps_per_epoch, n_epochs):
        wct = 0.0
        wct_per_step = 0
        assert 0 in epoch_wct_dict, "epoch_wct_dict must contain 0"
        for step in range(steps_per_epoch):
            if step in epoch_wct_dict:
                wct_per_step = epoch_wct_dict[step]
            wct += wct_per_step
        return wct * n_epochs

    def training_start(self, state: State, logger: Logger):
        self.wall_clock_train = 0.0
        self.original_max_epochs = state.max_epochs
        # maybe override epoch_list
        if self.all_epochs:
            self.epoch_list = list(range(state.max_epochs))
            log.info(f"all_epochs=True, overriding epoch_list to be every epoch from 0 to {state.max_epochs}")
        self.wct_dict = {e: {s: -1.0 for s in self.step_list} for e in self.epoch_list}
        state.max_epochs = len(self.epoch_list)

    def epoch_end(self, state: State, logger: Logger):
        prev_epoch = self.epoch_list[self.epoch_ix]
        epoch_wct_dict = self.wct_dict[prev_epoch]
        self.epoch_ix += 1
        if self.epoch_ix < len(self.epoch_list):
            next_epoch = self.epoch_list[self.epoch_ix]
        else:
            next_epoch = self.original_max_epochs

        state.epoch = next_epoch - 1
        state.step = next_epoch * state.steps_per_epoch
        n_epochs = next_epoch - prev_epoch

        self.wall_clock_train += self._compute_elapsed_wct(epoch_wct_dict, state.steps_per_epoch, n_epochs)
        logger.metric_epoch({'wall_clock_train': self.wall_clock_train})

    def batch_start(self, state: State, logger: Logger):
        if self.current_time is None:
            self.current_time = time.time()
            self.profile_examples = 0
            self.profile_steps = 0
            self.profile_time = 0.0

    def batch_end(self, state: State, logger: Logger):
        if self.current_time is not None:
            now = time.time()
            elapsed = now - self.current_time
            self.current_time = now
            self.profile_examples += state.last_batch_size * state.world_size
            self.profile_steps += 1
            self.profile_time += elapsed

            if self.profile_steps >= self.min_steps:
                avg_throughput = self.profile_examples / self.profile_time
                avg_time_per_step = self.profile_time / self.profile_steps
                profile_epoch = self.epoch_list[self.epoch_ix]
                profile_step = self.step_list[self.step_ix]
                self.wct_dict[profile_epoch][profile_step] = avg_time_per_step
                logger.metric_batch({'throughput/step': avg_throughput})

                self.current_time = None
                self.step_ix += 1
                if self.step_ix == len(self.step_list):
                    self.step_ix = 0
                    raise BreakEpochException
                else:
                    state.step = state.epoch * state.steps_per_epoch + self.step_list[self.step_ix]


@dataclass
class TimingMonitorHparams(CallbackHparams):
    min_steps: int = hp.optional(
        doc="minimum number of steps to use for measuring throughput",
        default=50,
    )
    epoch_list: List[int] = hp.optional(
        doc="list of epochs at which to measure throughput",
        default_factory=lambda: [0, 1],
    )
    step_list: List[int] = hp.optional(
        doc="list of steps at which to measure throughput",
        default_factory=lambda: [0, 50],
    )
    all_epochs: bool = hp.optional(
        doc="if true, override epoch_list and profile at all epochs.",
        default=False,
    )

    def initialize_object(self) -> TimingMonitor:
        return TimingMonitor(
            min_steps=self.min_steps,
            epoch_list=self.epoch_list,
            step_list=self.step_list,
            all_epochs=self.all_epochs,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Dict, Type
from unittest import mock

import pytest
import torch
from torch.optim.lr_scheduler import ExponentialLR, MultiStepLR, StepLR

from composer.core.types import ModelParameters
from composer.optim.pytorch_future import WarmUpLR
from composer.optim.scheduler import (ComposedScheduler, ConstantLRHparams, CosineAnnealingLRHparams,
                                      CosineAnnealingWarmRestartsHparams, ExponentialLRHparams, MultiStepLRHparams,
                                      SchedulerHparams, StepLRHparams, WarmUpLRHparams)
from composer.trainer.trainer_hparams import scheduler_registry

# for testing, we provide values for required hparams fields
schedulers: Dict[Type[SchedulerHparams], SchedulerHparams] = {
    StepLRHparams: StepLRHparams(step_size="5ep",),
    MultiStepLRHparams: MultiStepLRHparams(milestones=["5ep", "10ep"],),
    ExponentialLRHparams: ExponentialLRHparams(gamma=0.5,),
    CosineAnnealingLRHparams: CosineAnnealingLRHparams(T_max="1000ep",),
    CosineAnnealingWarmRestartsHparams: CosineAnnealingWarmRestartsHparams(T_0="1000ep",),
    WarmUpLRHparams: WarmUpLRHparams(),
    ConstantLRHparams: ConstantLRHparams()
}

time_field: Dict[Type[SchedulerHparams], str] = {
    StepLRHparams: 'step_size',
    MultiStepLRHparams: '',
    ExponentialLRHparams: '',
    CosineAnnealingLRHparams: 'T_max',
    CosineAnnealingWarmRestartsHparams: 'T_0',
    WarmUpLRHparams: 'warmup_iters',
    ConstantLRHparams: ''
}

EXPECTED_RESULTS_TIME_CONVERSION = {
    '33ep12ba': {
        'steps': 3312,
        'epochs': 33
    },
    '17ep': {
        'steps': 1700,
        'epochs': 17
    },
    '5050ba': {
        'steps': 5050,
        'epochs': 50
    },
    42: {
        'steps': 42,
        'epochs': 42
    }
}


@pytest.fixture
def dummy_parameters() -> ModelParameters:
    net = torch.nn.Sequential(torch.nn.Linear(5, 2), torch.nn.ReLU())
    return net.parameters()


@pytest.fixture
def dummy_optimizer(dummy_parameters) -> torch.optim.Optimizer:
    return torch.optim.SGD(dummy_parameters, 0.1)


@pytest.mark.parametrize("scheduler_name", scheduler_registry.keys())
class TestSchedulerInit():

    def test_scheduler_initialization(self, scheduler_name, dummy_optimizer):

        # create the scheduler hparams object
        obj: Type[SchedulerHparams] = scheduler_registry[scheduler_name]
        scheduler_hparams = schedulers[obj]

        # create the scheduler object using the hparams
        scheduler, interval = scheduler_hparams.initialize_object(dummy_optimizer, steps_per_epoch=1)
        assert isinstance(scheduler, scheduler_hparams.scheduler_object)  # type: ignore
        assert interval == scheduler_hparams.interval  # type: ignore

    @pytest.mark.parametrize('timestrings', EXPECTED_RESULTS_TIME_CONVERSION.keys())
    @pytest.mark.parametrize('interval', ['steps', 'epochs'])
    def test_scheduler_time_conversion(self, scheduler_name, dummy_optimizer, timestrings, interval):
        expected = EXPECTED_RESULTS_TIME_CONVERSION[timestrings][interval]
        obj: Type[SchedulerHparams] = scheduler_registry[scheduler_name]

        if time_field[obj]:
            scheduler_hparams = schedulers[obj]
            with mock.patch.object(scheduler_hparams, time_field[obj], timestrings), \
                mock.patch.object(scheduler_hparams, 'interval', interval):

                scheduler, interval = scheduler_hparams.initialize_object(dummy_optimizer, steps_per_epoch=100)

                assert getattr(scheduler, time_field[obj]) == expected


@pytest.fixture
def optimizer(dummy_model):
    return torch.optim.SGD(dummy_model.parameters(), lr=1)


class TestComposedScheduler():

    def _test(self, scheduler, targets, epochs, optimizer, interval='epoch'):
        for epoch in range(epochs):
            for param_group, target in zip(optimizer.param_groups, targets):
                torch.testing.assert_allclose(target[epoch], param_group['lr'])
            optimizer.step()
            scheduler.step(interval)

    def test_composed(self, optimizer):
        epochs = 9
        targets = [[1 * 0.2 for x in range(4)] + [1 * 0.9**x for x in range(7)]]
        schedulers = [
            ExponentialLR(optimizer, gamma=0.9),
            WarmUpLR(optimizer, warmup_factor=0.2, warmup_iters=4, warmup_method="constant")
        ]
        schedulers = [(s, 'epoch') for s in schedulers]
        scheduler = ComposedScheduler(schedulers)
        self._test(scheduler, targets, epochs, optimizer)

    def test_composed_linear(self, optimizer):
        epochs = 9
        targets = [[1 * 0.5 + (x/4 * 0.5) for x in range(4)] + [1 * 0.9**x for x in range(2)] + \
                   [1 * 0.9**x for x in range(2, 7)]]
        schedulers = [
            ExponentialLR(optimizer, gamma=0.9),
            WarmUpLR(optimizer, warmup_factor=0.5, warmup_iters=4, warmup_method="linear")
        ]
        schedulers = [(s, 'epoch') for s in schedulers]
        scheduler = ComposedScheduler(schedulers)
        self._test(scheduler, targets, epochs, optimizer)

    def test_composed_linear2(self, optimizer):
        epochs = 9
        targets = [[1 * 0.5 + (x/4 * 0.5) for x in range(4)] + \
                   [1 * 0.9**x for x in range(2)] + [1 * 0.9**x * 0.1 for x in range(2, 7)]]
        schedulers = [
            ExponentialLR(optimizer, gamma=0.9),
            MultiStepLR(optimizer, milestones=[6], gamma=0.1),
            WarmUpLR(optimizer, warmup_factor=0.5, warmup_iters=4, warmup_method="linear")
        ]
        schedulers = [(s, 'epoch') for s in schedulers]
        scheduler = ComposedScheduler(schedulers)
        self._test(scheduler, targets, epochs, optimizer)

    def test_composed_linear_from_zero(self, optimizer):
        epochs = 9
        targets = [[1 * 0.0 + (x / 4 * 1.0) for x in range(4)] + [1 * 0.9**x for x in range(7)]]
        schedulers = [
            ExponentialLR(optimizer, gamma=0.9),
            WarmUpLR(optimizer, warmup_factor=0, warmup_iters=4, warmup_method="linear")
        ]
        schedulers = [(s, 'epoch') for s in schedulers]
        scheduler = ComposedScheduler(schedulers)
        self._test(scheduler, targets, epochs, optimizer)

    def test_composed_linear_from_zero_step(self, optimizer):
        epochs = 9
        targets = [[x / 4 for x in range(4)] + [1.0 for _ in range(7)]]
        schedulers = [
            (ExponentialLR(optimizer, gamma=0.9), 'epoch'),  # should never trigger
            (WarmUpLR(optimizer, warmup_factor=0, warmup_iters=4, warmup_method="linear"), 'batch')
        ]
        scheduler = ComposedScheduler(schedulers)
        self._test(scheduler, targets, epochs, optimizer, interval='batch')

    @pytest.mark.xfail
    def test_validate_compose_multistep(self, optimizer):
        schedulers = [
            ExponentialLR(optimizer, gamma=0.9),
            WarmUpLR(optimizer, warmup_factor=0, warmup_iters=4, warmup_method="linear"),
            MultiStepLR(optimizer, milestones=[3], gamma=0.1)
        ]
        schedulers = [(s, 'epoch') for s in schedulers]

        with pytest.raises(ValueError):
            ComposedScheduler(schedulers)

    @pytest.mark.xfail
    def test_validate_compose_step(self, optimizer):
        schedulers = [
            ExponentialLR(optimizer, gamma=0.9),
            WarmUpLR(optimizer, warmup_factor=0, warmup_iters=4, warmup_method="linear"),
            StepLR(optimizer, step_size=2, gamma=0.1)
        ]
        schedulers = [(s, 'epoch') for s in schedulers]

        with pytest.raises(ValueError):
            ComposedScheduler(schedulers)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Optional, Tuple

import torch
from torch import Tensor
from torchmetrics.classification.accuracy import Accuracy
from torchmetrics.collections import MetricCollection

from composer.core.types import BatchPair
from composer.models.loss import CrossEntropyLoss, soft_cross_entropy

if TYPE_CHECKING:
    from composer.core.types import Batch, Metrics, Tensors


class BaseMosaicModel(torch.nn.Module, abc.ABC):

    @abc.abstractmethod
    def loss(self, outputs: Any, batch: Batch, *args, **kwargs) -> Tensors:
        """Compute the loss of the model.

        Args:
            outputs: output of the foward pass
            batch: input batch from dataloader

        Returns:
            The loss as a Tensors object.
        """
        pass

    @abc.abstractmethod
    def forward(self, batch: Batch) -> Tensors:
        """Compute model output given an input.

        Args:
            batch (Batch): input batch for forward pass

        Returns:
            Tensors: result that is passed to loss
        """
        pass

    @abc.abstractmethod
    def metrics(self, train: bool = False) -> Metrics:
        """Used to get metrics that will be used to evaluate the model. Note
        that each metric keeps states which are updated with data seen so far.
        As a result, different metric objects should be used for training and validation.
        See: https://torchmetrics.readthedocs.io/en/latest/pages/overview.html
        for more details.

        Args:
            train (optional): True to return metrics that should be computed during training
            and False otherwise. Default is False.

        Returns:
            Either a torchmetrics.Metric object or a torchmetrics.MetricCollection
            object.
        """
        pass

    @abc.abstractmethod
    def validate(self, batch: Batch) -> Tuple[Any, Any]:
        """
        Compute model outputs given data. The output of this function
        will be used as input to any metrics being computed for the model.

        Args:
            batch: The data to perform validation with. Specified as a tuple of
            tensors (input, target).

        Returns:
            Tuple[Any, Any]: tuple that is passed directly
                             to the provided torchmetrics.Metric's update method.
        """
        pass


class MosaicClassifier(BaseMosaicModel):

    num_classes: Optional[int] = None

    def __init__(self, module: torch.nn.Module) -> None:
        super().__init__()
        self.train_acc = Accuracy()
        self.val_acc = Accuracy()
        self.val_loss = CrossEntropyLoss()
        self.module = module

        # TODO(issue #249): Needs to be replaced
        if hasattr(self.module, "num_classes"):
            self.num_classes = getattr(self.module, "num_classes")

    def loss(self, outputs: Any, batch: BatchPair, *args, **kwargs) -> Tensors:
        _, y = batch
        assert isinstance(outputs, Tensor), "Loss expects outputs as Tensor"
        assert isinstance(y, Tensor), "Loss does not support multiple target Tensors"
        return soft_cross_entropy(outputs, y, *args, **kwargs)

    def metrics(self, train: bool = False) -> Metrics:
        return self.train_acc if train else MetricCollection([self.val_acc, self.val_loss])

    def forward(self, batch: BatchPair) -> Tensor:
        x, y = batch
        logits = self.module(x)

        return logits

    def validate(self, batch: BatchPair) -> Tuple[Any, Any]:
        assert self.training is False, "For validation, model must be in eval mode"
        inputs, targets = batch
        logits = self.forward(batch)
        return logits, targets



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.resnet56_cifar10.model import CIFAR10_ResNet56 as CIFAR10_ResNet56
from composer.models.resnet56_cifar10.resnet56_cifar10_hparams import CIFARResNetHparams as CIFARResNetHparams

_task = 'Image Classification'
_dataset = 'CIFAR10'
_name = 'ResNet56'
_quality = ''
_metric = 'Top-1 Accuracy'
_ttt = '?'
_hparams = 'hparams.yaml'



# Copyright 2021 MosaicML. All Rights Reserved.

import glob
import os
import random
from dataclasses import dataclass

import numpy as np
import torch
import torchvision
import yahp as hp
from torch.utils.data import Dataset

from composer.datasets.hparams import DataloaderSpec, DatasetHparams

PATCH_SIZE = [1, 192, 160]


def my_collate(batch):
    """Custom collate function to handle images with different depths.

    """
    data = [item[0] for item in batch]
    target = [item[1] for item in batch]

    return [torch.Tensor(data), torch.Tensor(target)]


@dataclass
class BratsDatasetHparams(DatasetHparams):
    is_train: bool = hp.required("whether to load the training or validation dataset")
    datadir: str = hp.required("data directory")
    download: bool = hp.required("whether to download the dataset, if needed")
    drop_last: bool = hp.optional("Whether to drop the last samples for the last batch", default=True)
    shuffle: bool = hp.optional("Whether to shuffle the dataset for each epoch", default=True)
    oversampling: float = hp.optional("oversampling", default=0.33)

    def initialize_object(self) -> DataloaderSpec:

        datadir = self.datadir
        oversampling = self.oversampling

        x_train, y_train, x_val, y_val = get_data_split(datadir)
        train_dataset = PytTrain(x_train, y_train, oversampling)
        val_dataset = PytVal(x_val, y_val)
        if self.is_train:
            return DataloaderSpec(
                dataset=train_dataset,
                drop_last=self.drop_last,
                shuffle=self.shuffle,
            )
        else:
            return DataloaderSpec(
                dataset=val_dataset,
                drop_last=self.drop_last,
                shuffle=self.shuffle,
                collate_fn=my_collate,  # type: ignore
            )


def coin_flip(prob):
    return random.random() < prob


def random_augmentation(probability, augmented, original):
    condition = coin_flip(probability)
    neg_condition = condition ^ True
    return condition * augmented + neg_condition * original


class Crop(object):

    def __call__(self, data, oversampling):
        img, lbl = data["image"], data["label"]

        def randrange(max_range):
            return 0 if max_range == 0 else random.randrange(max_range)

        def get_cords(cord, idx):
            return cord[idx], cord[idx] + PATCH_SIZE[idx]

        def _rand_crop(image, label):
            ranges = [s - p for s, p in zip(image.shape[1:], PATCH_SIZE)]

            cord = [randrange(x) for x in ranges]
            low_x, high_x = get_cords(cord, 0)
            low_y, high_y = get_cords(cord, 1)
            image = image[:, low_x:high_x, low_y:high_y]
            label = label[:, low_x:high_x, low_y:high_y]
            return image, label, [low_x, high_x, low_y, high_y]

        def rand_foreg_cropd(image, label):

            import scipy
            cl = np.random.choice(np.unique(label[label > 0]))
            foreg_slices = scipy.ndimage.find_objects(scipy.ndimage.measurements.label(label == cl)[0])
            foreg_slices = [x for x in foreg_slices if x is not None]
            slice_volumes = [np.prod([s.stop - s.start for s in sl]) for sl in foreg_slices]
            slice_idx = np.argsort(slice_volumes)[-2:]
            foreg_slices = [foreg_slices[i] for i in slice_idx]
            if not foreg_slices:
                return _rand_crop(image, label)
            foreg_slice = foreg_slices[random.randrange(len(foreg_slices))]
            low_x, high_x = adjust(foreg_slice, PATCH_SIZE, label, 1)
            low_y, high_y = adjust(foreg_slice, PATCH_SIZE, label, 2)
            image = image[:, low_x:high_x, low_y:high_y]
            label = label[:, low_x:high_x, low_y:high_y]
            return image, label, [low_x, high_x, low_y, high_y]

        def adjust(foreg_slice, patch_size, label, idx):

            diff = patch_size[idx - 1] - (foreg_slice[idx].stop - foreg_slice[idx].start)
            sign = -1 if diff < 0 else 1
            diff = abs(diff)
            ladj = randrange(diff)
            hadj = diff - ladj
            low = max(0, foreg_slice[idx].start - sign * ladj)
            high = min(label.shape[idx], foreg_slice[idx].stop + sign * hadj)
            diff = patch_size[idx - 1] - (high - low)
            if diff > 0 and low == 0:
                high += diff
            elif diff > 0:
                low -= diff
            return low, high

        if random.random() < oversampling:
            img, lbl, _ = rand_foreg_cropd(img, lbl)
        else:
            img, lbl, _ = _rand_crop(img, lbl)

        return {'image': img, 'label': lbl}


class Noise(object):

    def __call__(self, data, oversampling):
        img, lbl = data["image"], data["label"]
        std = np.random.uniform(0.0, oversampling)
        noise = np.random.normal(0, scale=std, size=img.shape)
        img_noised = img + noise
        img = random_augmentation(0.15, img_noised, img)

        return {'image': img, 'label': lbl}


class Blur(object):

    def __call__(self, data):
        img, lbl = data["image"], data["label"]

        transf = torchvision.transforms.GaussianBlur(kernel_size=3, sigma=(0.5, 1.5))
        img_blured = transf(torch.Tensor(img)).numpy()
        img = random_augmentation(0.15, img_blured, img)

        return {'image': img, 'label': lbl}


class Brightness(object):

    def __call__(self, data):
        img, lbl = data["image"], data["label"]
        brightness_scale = random_augmentation(0.15, np.random.uniform(0.7, 1.3), 1.0)
        img = img * brightness_scale

        return {'image': img, 'label': lbl}


class Contrast(object):

    def __call__(self, data):
        img, lbl = data["image"], data["label"]
        min_, max_ = np.min(img), np.max(img)
        scale = random_augmentation(0.15, np.random.uniform(0.65, 1.5), 1.0)

        img = torch.clamp(torch.Tensor(img * scale), min_, max_).numpy()
        return {'image': img, 'label': lbl}


class Flips(object):

    def __call__(self, data):
        img, lbl = data["image"], data["label"]
        axes = [1, 2]
        prob = 1 / len(axes)

        for axis in axes:
            if random.random() < prob:
                img = np.flip(img, axis=axis).copy()
                lbl = np.flip(lbl, axis=axis).copy()

        return {'image': img, 'label': lbl}


class Transpose(object):

    def __call__(self, data):
        img, lbl = data["image"], data["label"]
        img, lbl = img.transpose((1, 0, 2, 3)), lbl.transpose((1, 0, 2, 3))

        return {'image': img, 'label': lbl}


class PytTrain(Dataset):

    def __init__(self, images, labels, oversampling, transform=None):
        self.images, self.labels = images, labels
        self.oversampling = oversampling
        self.transform = transform
        self.rand_crop = Crop()
        self.transpose = Transpose()
        self.contrast = Contrast()
        self.noise = Noise()
        self.blur = Blur()
        self.flips = Flips()
        self.bright = Brightness()

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        data = {"image": np.load(self.images[idx]), "label": np.load(self.labels[idx])}
        data = self.rand_crop(data, self.oversampling)
        data = self.flips(data)
        data = self.noise(data, self.oversampling)
        data = self.blur(data)
        data = self.bright(data)
        data = self.contrast(data)
        data = self.transpose(data)

        return data["image"], data["label"]


class PytVal(Dataset):

    def __init__(self, images, labels):
        self.images, self.labels = images, labels

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        data = {"image": np.load(self.images[idx]), "label": np.load(self.labels[idx])}
        return data["image"], data["label"]


def load_data(path, files_pattern):
    data = sorted(glob.glob(os.path.join(path, files_pattern)))
    assert len(data) > 0, f"Found no data at {path}"
    return data


def get_split(data, idx):
    return list(np.array(data)[idx])


def get_data_split(path: str):
    from sklearn.model_selection import KFold

    val_cases_list = []
    kfold = KFold(n_splits=5, shuffle=True, random_state=0)
    imgs = load_data(path, "*_x.npy")
    lbls = load_data(path, "*_y.npy")
    assert len(imgs) == len(lbls), f"Found {len(imgs)} volumes but {len(lbls)} corresponding masks"
    train_imgs, train_lbls, val_imgs, val_lbls = [], [], [], []

    train_idx, val_idx = list(kfold.split(imgs))[0]
    train_imgs = get_split(imgs, train_idx)
    train_lbls = get_split(lbls, train_idx)
    val_imgs = get_split(imgs, val_idx)
    val_lbls = get_split(lbls, val_idx)

    return train_imgs, train_lbls, val_imgs, val_lbls



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State
from composer.core.types import Model, Optimizers

log = logging.getLogger(__name__)


@dataclass
class LayerFreezingHparams(AlgorithmHparams):

    freeze_start: float = hp.optional(doc='The percentage of epochs to run before freezing begins.', default=0.5)
    freeze_level: float = hp.optional(doc='Scale factor for the percentage of the network to freeze.', default=1.0)

    def initialize_object(self) -> LayerFreezing:
        return LayerFreezing(**asdict(self))


def _freeze_schedule(current_epoch, max_epochs, freeze_start, freeze_level):
    """Implements a linear schedule for freezing.

    The schedule is linear and begins with no freezing and
    linearly increases the fraction of layers frozen, reaching
    the fraction specified by 'freeze_level' at the final epoch.
    The start of freezing is given as a fraction of the total
    number of epochs, and is set with 'freeze_start'.

    Args:
        current_epoch: Integer specifying the current epoch.
        max_epochs: The max number of epochs training will run for.
        freeze_start: The fraction of epochs to run before freezing begins.
        freeze_level: The maximum fraction of levels to freeze.
    """
    # Calculate the epoch to start freezing
    freeze_start_epoch = int(freeze_start * max_epochs)
    # No freezing if the current epoch is less than this
    if current_epoch <= freeze_start_epoch:
        freeze_percentage = 0
    else:
        # Calculate the total time for freezing to occur
        reduced_time = max_epochs - freeze_start_epoch
        # Calculate the amount of freezing time that has elapsed
        time_elapsed = current_epoch - freeze_start_epoch
        # Calculate the fraction of the freezing time elapsed.
        scaling = time_elapsed / reduced_time
        # Scale this fraction by the amount of freezing to do.
        freeze_percentage = freeze_level * scaling

    return freeze_percentage


def _get_layers(module, flat_children):
    """Helper function to get all submodules.

    Does a depth first search to flatten out modules which
    contain parameters.

    Args:
        module: Current module to search.
        flat_children: List containing modules.
    """
    # Check if given module has no children and parameters.
    if (len(list(module.children())) == 0 and len(list(module.parameters())) > 0):
        flat_children.append(module)
    else:
        # Otherwise, continue the search over its children.
        for child in module.children():
            _get_layers(child, flat_children)


def _remove_param_from_optimizers(p, optimizers):
    """Helper function to freeze the training of a parameter.

    To freeze a parameter, it must be removed from the optimizer,
    otherwise momentum and weight decay may still be applied.

    Args:
        p: The parameter being frozen.
        optimizers: The optimizers used during training.
    """
    # Force optimizers to be iterable
    if not isinstance(optimizers, (list, tuple)):
        optimizers = [optimizers]

    # Search over params in the optimizers to find and remove the
    # given param. Necessary due to the way params are stored.
    for optimizer in optimizers:
        for group in optimizer.param_groups:
            group['params'] = list(filter(lambda x: id(x) != id(p), group['params']))


def freeze_layers(
    model: Model,
    optimizers: Optimizers,
    current_epoch: int,
    max_epochs: int,
    freeze_start: float,
    freeze_level: float,
    logger: Logger,
):
    """
    Implements the layer freezing algorithm. During training, progressively freeze the
    layers of the network starting with the earlier layers.

    Args:
        model: An instance of the model being trained.
        optimizers: The optimizers used during training.
        current_epoch: Integer specifying the current epoch.
        max_epochs: The max number of epochs training will run for.
        freeze_start: The fraction of epochs to run before freezing begins.
        freeze_level: The maximum fraction of levels to freeze.
    """
    # Flatten out the layers
    flat_children = []
    _get_layers(model, flat_children)
    # Determine how many layers to freeze
    freeze_percentage = _freeze_schedule(current_epoch=current_epoch,
                                         max_epochs=max_epochs,
                                         freeze_start=freeze_start,
                                         freeze_level=freeze_level)
    freeze_depth = int(freeze_percentage * len(flat_children[0:-1]))

    # Freeze the parameters in the chosen layers
    for i, child in enumerate(flat_children[0:-1]):
        if i < freeze_depth:
            for p in child.parameters():
                _remove_param_from_optimizers(p, optimizers)
                # Do not compute gradients for this param.
                p.requires_grad = False

    # Log results
    log.info(f'Applied Layer Freezing'
             f' with freeze_start={freeze_start}, '
             f'freeze_level={freeze_level}. '
             f'Froze {freeze_depth} layers in the model which'
             f' equates to {freeze_percentage * 100}% of all layers.')
    logger.metric_epoch({
        'layer_freezing/layers_frozen': freeze_depth,
        'layer_freezing/percentage_frozen': freeze_percentage
    })

    return model


class LayerFreezing(Algorithm):
    """
    Algorithm to apply Layer Freezing to the model. Runs on Event.EPOCH_END.
    During training, progressively freeze the layers of the network starting
    with the earlier layers. Freezing starts after the percent of epochs
    specified by freeze_start have run. The fraction of layers increases
    linearly until it reaches freeze_level at the final epoch.

    Args:
        freeze_start: The fraction of epochs to run before freezing begins.
        freeze_level: The maximum fraction of levels to freeze.
    """

    def __init__(self, freeze_start: float = 0.5, freeze_level: float = 1.0):
        """
        __init__ is constructed from the same fields as in hparams.
        """
        self.hparams = LayerFreezingHparams(freeze_start, freeze_level)

    @property
    def find_unused_parameters(self) -> bool:
        """
        Override in order to tell DDP that some parameters will not
        have gradients computed for them after layer freezing is applied.
        """
        return True

    def match(self, event: Event, state: State) -> bool:
        """
        Run Layer Freezing on epoch end.
        """
        return event == Event.EPOCH_END

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """
        Apply Layer Freezing.
        """
        optimizers = state.optimizers
        assert optimizers is not None
        state.model = freeze_layers(
            model=state.model,
            optimizers=optimizers,
            current_epoch=state.epoch,
            max_epochs=state.max_epochs,
            freeze_start=self.hparams.freeze_start,
            freeze_level=self.hparams.freeze_level,
            logger=logger,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExcite as SqueezeExcite
from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExcite2d as SqueezeExcite2d
from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExciteConv2d as SqueezeExciteConv2d
from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExciteHparams as SqueezeExciteHparams
from composer.algorithms.squeeze_excite.squeeze_excite import apply_se as apply_se

_name = 'SqueezeExcite'
_class_name = 'SqueezeExcite'
_functional = 'apply_se'
_tldr = 'Replaces eligible layers with Squeeze-Excite layers'
_attribution = 'Hu et al, 2017'
_link = 'https://arxiv.org/abs/1709.01507'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import logging
import os
import random
import warnings
from typing import Any, Dict, Optional

import numpy as np
import torch
import yaml

from composer.core import Event, State
from composer.core.types import StateDict
from composer.trainer.ddp import DDP
from composer.trainer.devices.device import Device
from composer.utils import seed_all

log = logging.getLogger(__name__)


class CheckpointLoader:
    """Manager for initializing state and restoring RNG state from existing checkpoints.
    
    Args:
        checkpoint_filepath (str): The path to an existing checkpoint file.
    """

    def __init__(self, checkpoint_filepath: str):
        self.state_dict = torch.load(checkpoint_filepath, map_location='cpu')

    def load_checkpoint(self, state: State):
        """Initialize state from the loaded checkpoint's data.
        """

        state.load_state_dict(self.state_dict["state"])
        self.checkpoint_rng_state = self._get_checkpoint_rng_state(state, self.state_dict["rng"])

        if state.seed is not None:
            seed_all(state.seed)

    def restore_checkpoint_rng_state(self, state: State, device: Device):
        """Restore the state of all RNG objects in this context from the loaded checkpoint's data.
        """

        if self.checkpoint_rng_state is None:
            return

        assert state.world_size == len(
            self.checkpoint_rng_state['torch']
        ), f"invariant violation: if the rng state is being restored, then" \
            "the world size should be the same as in the checkpoint."

        torch.set_rng_state(self.checkpoint_rng_state['torch'][state.global_rank])
        device.load_state_dict(self.checkpoint_rng_state['device'][state.global_rank])
        random.setstate(self.checkpoint_rng_state['python'][state.global_rank])
        np.random.set_state(self.checkpoint_rng_state['numpy'][state.global_rank])

        self.checkpoint_rng_state = None

    def _get_checkpoint_rng_state(self, state: State, checkpoint_rng_state: StateDict) -> Optional[StateDict]:
        original_world_size = len(checkpoint_rng_state["torch"])
        if original_world_size == state.world_size:
            return checkpoint_rng_state
        else:
            warnings.warn(f"The checkpoint was created with world_size({original_world_size}), "
                          f"which differs from the current world_size({state.world_size})."
                          f"RNG state will not be restored.")


class Checkpointer:
    """Manager for saving state to checkpoint files.
    
    Args:
        checkpoint_folder (str): The path to the folder to store checkpoints in.
        checkpoint_interval (int): The amount of time units to wait between checkpoints.
        checkpoint_interval_unit (str): The unit (`"ep"` or `"it"`) that `checkpoint_interval` should be measured in.
    """

    def __init__(self, checkpoint_folder: str, checkpoint_interval: int, checkpoint_interval_unit: str):
        if checkpoint_interval_unit.lower() == "ep":
            self.save_event = Event.EPOCH_END
        elif checkpoint_interval_unit.lower() == "it":
            self.save_event = Event.BATCH_END
        else:
            raise ValueError(f"Unknown checkpointing interval: {checkpoint_interval_unit}")
        self.checkpoint_folder = checkpoint_folder
        self.save_interval = checkpoint_interval

    def should_checkpoint(self, state: State, event: Event) -> bool:
        """Given the current state and event, determine whether a checkpoint needs to be created.

        Args:
            state (State): The current State of the trainer.
            event (Event): The current Event being executed.
        """

        if event != self.save_event:
            return False
        if self.save_event == Event.EPOCH_END:
            return state.epoch % self.save_interval == 0
        if self.save_event == Event.BATCH_END:
            return state.step % self.save_interval == 0
        return False

    def save_checkpoint(self, state: State, device: Device, ddp: DDP, config: Optional[Dict[str, Any]] = None) -> None:
        """Save the current state to a a new checkpoint file.

        Args:
            state (State): The current State of the trainer.
            device (Device): The Device in use by this process.
            ddp (DDP): The DDP engine in use by this trainer.
            config (Optional[Dict[str, Any]]): The hparams used to initialize this trainer, if any.
        """

        state_dict = {
            'state': state.state_dict(),  # should be the same across all ranks. per-rank state not stored
            'rng': self._get_rng_state(device=device, ddp=ddp),  # stored across all ranks
        }
        if not state.is_rank_zero:
            # only rank 0 saves checkpoints
            # Need the check down here so all the DDP syncs will work for generating the checkpoint
            return

        # The trainer will only have _hparams_yaml set if it is instantiated with create_from_hparams
        if config:
            hparams_path = os.path.join(self.checkpoint_folder, "hparams.yaml")
            os.makedirs(self.checkpoint_folder, mode=0o775, exist_ok=True)
            config_yaml_str = yaml.dump(config)
            try:
                with open(hparams_path, "x") as f:
                    # Storing the hparams in a separate file so they can be modified before resuming
                    f.write(config_yaml_str)
            except FileExistsError as e:
                with open(hparams_path, "r") as f:
                    # comparing the parsed hparams to ignore whitespace and formatting differences
                    if yaml.safe_load(config_yaml_str) != yaml.safe_load(f):
                        raise RuntimeError(f"The hparams in the existing checkpoint folder {self.checkpoint_folder} "
                                           "differ from those being used in the current training run. "
                                           "Please specify a new checkpoint folder.") from e
        if self.save_event == Event.EPOCH_END:
            filename = f"ep{state.epoch}.pt"
        elif self.save_event == Event.BATCH_END:
            filename = f"it{state.step}.pt"
        else:
            raise ValueError(f"Invalid checkpoint event: {self.save_event}")
        save_file = os.path.join(self.checkpoint_folder, filename)
        with open(save_file, 'xb') as f:
            torch.save(state_dict, f)
        log.info(f'Trainer checkpoint saved to {save_file}')

    def _get_rng_state(self, device: Device, ddp: DDP) -> StateDict:
        rng_state = {
            "python": ddp.all_gather_object(random.getstate()),
            "numpy": ddp.all_gather_object(np.random.get_state()),
            "torch": ddp.all_gather_object(torch.random.get_rng_state()),
            "device": ddp.all_gather_object(device.state_dict()),
        }
        # casting the state dict as on non-rank-0, entries will be None-like
        return rng_state



# Copyright 2021 MosaicML. All Rights Reserved.

# type: ignore
# yapf: disable
# Code original from TODO CITATION.
from functools import partial
from typing import Callable, List, Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.models import ResNet
from torchvision.models.resnet import conv1x1, conv3x3

from composer.models import Initializer


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None) -> None:
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Torchvision_ResNet(ResNet):

    def __init__(self, block, layers, num_classes=1000, width=64):
        """To make it possible to vary the width, we need to
        override the constructor of the torchvision resnet."""

        nn.Module.__init__(self)  # Skip the parent constructor. This replaces it.
        self._norm_layer = nn.BatchNorm2d
        self.inplanes = width
        self.dilation = 1
        self.groups = 1
        self.base_width = 64

        # The initial convolutional layer.
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = self._norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        # The subsequent blocks.
        self.layer1 = self._make_layer(block, width, layers[0])
        self.layer2 = self._make_layer(block, width * 2, layers[1], stride=2, dilate=False)
        self.layer3 = self._make_layer(block, width * 4, layers[2], stride=2, dilate=False)
        self.layer4 = self._make_layer(block, width * 8, layers[3], stride=2, dilate=False)

        # The last layers.
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(width * 8 * block.expansion, num_classes)

        # Default init.
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)


class ImageNet_ResNet(nn.Module):
    """A residual neural network as originally designed for ImageNet."""

    def __init__(self, model_fn, initializers: List[Union[str, Initializer]], outputs=None):
        super(ImageNet_ResNet, self).__init__()

        # TODO(issue #249):
        self.num_classes = outputs
        self.model = model_fn(num_classes=outputs or 1000)
        self.criterion = nn.CrossEntropyLoss()
        for initializer in initializers:
            initializer = Initializer(initializer)
            self.apply(initializer.get_initializer())

    def forward(self, x):
        return self.model(x)

    @staticmethod
    def is_valid_model_name(model_name):
        return (model_name.startswith('imagenet_resnet_') and 4 >= len(model_name.split('_')) >= 3 and
                model_name.split('_')[2].isdigit() and int(model_name.split('_')[2]) in [18, 34, 50, 101, 152, 200])

    @staticmethod
    def get_model_from_name(model_name, initializers, outputs=1000):
        """Name: imagenet_resnet_D[_W].
        D is the model depth (e.g., 50 for ResNet-50).
        W is the model width - the number of filters in the first
        residual layers. By default, this number is 64."""

        model_arch = {
            18: {
                'block': BasicBlock,
                'layers': [2, 2, 2, 2]
            },
            34: {
                'block': BasicBlock,
                'layers': [3, 4, 6, 3]
            },
            50: {
                'block': Bottleneck,
                'layers': [3, 4, 6, 3]
            },
            101: {
                'block': Bottleneck,
                'layers': [3, 4, 23, 3]
            },
            152: {
                'block': Bottleneck,
                'layers': [3, 8, 36, 3]
            },
            200: {
                'block': Bottleneck,
                'layers': [3, 24, 36, 3]
            },
            269: {
                'block': Bottleneck,
                'layers': [3, 30, 48, 8]
            },
        }

        if not ImageNet_ResNet.is_valid_model_name(model_name):
            raise ValueError(f'Invalid model name: {model_name}')

        num = int(model_name.split('_')[2])
        block = model_arch[num]['block']
        layers = model_arch[num]['layers']
        model_fn = partial(Torchvision_ResNet, block, layers)
        if len(model_name.split('_')) == 4:
            width = int(model_name.split('_')[3])
            model_fn = partial(model_fn, width=width)

        return ImageNet_ResNet(model_fn, initializers, outputs)


class CIFAR_ResNet(nn.Module):
    """A residual neural network as originally designed for CIFAR-10."""

    class Block(nn.Module):
        """A ResNet block."""

        def __init__(self, f_in: int, f_out: int, downsample=False):
            super(CIFAR_ResNet.Block, self).__init__()

            stride = 2 if downsample else 1
            self.conv1 = nn.Conv2d(f_in, f_out, kernel_size=3, stride=stride, padding=1, bias=False)
            self.bn1 = nn.BatchNorm2d(f_out)
            self.conv2 = nn.Conv2d(f_out, f_out, kernel_size=3, stride=1, padding=1, bias=False)
            self.bn2 = nn.BatchNorm2d(f_out)
            self.relu = nn.ReLU(inplace=True)

            # No parameters for shortcut connections.
            if downsample or f_in != f_out:
                self.shortcut = nn.Sequential(
                    nn.Conv2d(f_in, f_out, kernel_size=1, stride=2, bias=False),
                    nn.BatchNorm2d(f_out),
                )
            else:
                self.shortcut = nn.Sequential()

        def forward(self, x):
            out = self.relu(self.bn1(self.conv1(x)))
            out = self.bn2(self.conv2(out))
            out += self.shortcut(x)
            return self.relu(out)

    def __init__(self, plan, initializers: List[Union[str, Initializer]], outputs=None):
        super(CIFAR_ResNet, self).__init__()
        outputs = outputs or 10

        # TODO(issue #249): Replace
        self.num_classes = outputs

        # Initial convolution.
        current_filters = plan[0][0]
        self.conv = nn.Conv2d(3, current_filters, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn = nn.BatchNorm2d(current_filters)
        self.relu = nn.ReLU(inplace=True)

        # The subsequent blocks of the ResNet.
        blocks = []
        for segment_index, (filters, num_blocks) in enumerate(plan):
            for block_index in range(num_blocks):
                downsample = segment_index > 0 and block_index == 0
                blocks.append(CIFAR_ResNet.Block(current_filters, filters, downsample))
                current_filters = filters

        self.blocks = nn.Sequential(*blocks)

        # Final fc layer. Size = number of filters in last segment.
        self.fc = nn.Linear(plan[-1][0], outputs)
        self.criterion = nn.CrossEntropyLoss()

        for initializer in initializers:
            initializer = Initializer(initializer)
            self.apply(initializer.get_initializer())

    def forward(self, x):
        out = self.relu(self.bn(self.conv(x)))
        out = self.blocks(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.fc(out)
        return out

    @staticmethod
    def is_valid_model_name(model_name):
        return (model_name.startswith('cifar_resnet_') and 4 >= len(model_name.split('_')) >= 3 and
                model_name.split('_')[2].isdigit() and int(model_name.split('_')[2]) in [56])

    @staticmethod
    def get_model_from_name(model_name, initializers: List[Initializer], outputs=10):
        """The naming scheme for a ResNet is 'cifar_resnet_D[_W]'.
        D is the model depth (e.g. cifar_resnet56)

        """

        if not CIFAR_ResNet.is_valid_model_name(model_name):
            raise ValueError('Invalid model name: {}'.format(model_name))

        depth = int(model_name.split('_')[2])
        if len(model_name.split('_')) == 3:
            width = 16
        else:
            width = int(model_name.split('_')[4])

        if (depth - 2) % 3 != 0:
            raise ValueError('Invalid CIFAR_ResNet depth: {}'.format(depth))
        num_blocks = (depth - 2) // 6

        model_arch = {
            56: [(width, num_blocks), (2 * width, num_blocks), (4 * width, num_blocks)],
        }

        return CIFAR_ResNet(model_arch[depth], initializers, outputs)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional

import torch

from composer.algorithms.algorithm_hparams import AlgorithmHparams
from composer.core.types import Algorithm, Event, Logger, State

log = logging.getLogger(__name__)


class ChannelsLast(Algorithm):
    """ ChannelsLast algorithm runs on Event.TRAINING_START and changes
    the memory format of the model to `torch.channels_last`. This
    algorithm has no hyperparameters.
    """

    def match(self, event: Event, state: State) -> bool:
        """ Runs on Event.TRAINING_START
        """
        return event == Event.TRAINING_START

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """ Changes the memory format of the model to `torch.channels_last`
        """
        assert state.model is not None, 'Channels Last cannot be applied to None'
        # TODO: Double check model is moved to cuda with device type
        state.model.to(memory_format=torch.channels_last)  # type: ignore

        log.info(f'Model {state.model.__class__.__name__} changed to channels_last format.')


@dataclass
class ChannelsLastHparams(AlgorithmHparams):
    """ ChannelsLast algorithm has no hyperparameters.
    """
    pass

    def initialize_object(self) -> ChannelsLast:
        return ChannelsLast()



# Copyright 2021 MosaicML. All Rights Reserved.

from unittest.mock import Mock

import pytest
import torch
from torch import nn

from composer.core import surgery


class RecursiveLinear(nn.Linear):

    def __init__(self, in_features, out_features):
        super().__init__(in_features, out_features)

        # submodule has modified out_features to prevent infinite recursion during test
        self.submodule = nn.Linear(in_features, out_features - 1)


class SimpleReplacementPolicy(nn.Module):
    """
    Bundle the model, replacement function, and validation into one class
    """

    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(in_features=16, out_features=32)
        self.fc2 = nn.Linear(in_features=32, out_features=10)

    @staticmethod
    def maybe_replace_linear(module, module_index):
        if module.out_features in (10, 9):
            return RecursiveLinear(module.in_features, module.out_features)
        return None

    def policy(self):
        return {nn.Linear: self.maybe_replace_linear}

    def validate_replacements(self, recurse_on_replacements):
        assert type(self.fc1) is nn.Linear
        assert type(self.fc2) is RecursiveLinear

        if recurse_on_replacements:
            assert type(self.fc2.submodule) is RecursiveLinear
            assert type(self.fc2.submodule.submodule) is nn.Linear
        else:
            assert type(self.fc2.submodule) is nn.Linear


class ModuleIdxReplacementPolicy(SimpleReplacementPolicy):
    """
    Test replacing only the first instance of a Linear layer
    """

    @staticmethod
    def maybe_replace_linear(module, module_index):
        if module_index == 0:
            return RecursiveLinear(module.in_features, module.out_features)
        return None

    def validate_replacements(self, recurse_on_replacements):
        assert type(self.fc1) is RecursiveLinear
        assert type(self.fc2) is nn.Linear
        assert type(self.fc1.submodule) is nn.Linear


class NoOpReplacementPolicy(SimpleReplacementPolicy):

    def policy(self):
        return {nn.Conv2d: Mock(side_effect=AssertionError('test should not match on this layer'))}

    def validate_replacements(self, recurse_on_replacements):
        assert type(self.fc1) is nn.Linear
        assert type(self.fc2) is nn.Linear


@pytest.mark.parametrize('recurse_on_replacements', [True, False])
@pytest.mark.parametrize('model', [
    SimpleReplacementPolicy,
    ModuleIdxReplacementPolicy,
    NoOpReplacementPolicy,
])
def test_module_replacement(model, recurse_on_replacements):
    model = model()
    surgery.replace_module_classes(
        model,
        model.policy(),
        recurse_on_replacements=recurse_on_replacements,
    )

    model.validate_replacements(recurse_on_replacements)


class _CopyLinear(torch.nn.Module):

    def __init__(self, in_features: int, out_features: int):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = torch.nn.Parameter(torch.empty((out_features, in_features)))
        self.bias = None

    @staticmethod
    def from_linear(module: torch.nn.Module, module_index: int = -1):
        ret = _CopyLinear(in_features=module.in_features, out_features=module.out_features)  # type: ignore
        with torch.no_grad():
            # new param object
            ret.weight.copy_(module.weight)  # type: ignore
            ret.bias = module.bias  # same param object
        return ret



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.no_op_model.no_op_model import NoOpModel as NoOpModel
from composer.algorithms.no_op_model.no_op_model import NoOpModelHparams as NoOpModelHparams

_name = 'No-Op Model'
_class_name = 'NoOpModel'
_functional = ''
_tldr = 'Replaces the model with a dummy model. Used for dataloader profiling.'
_attribution = 'MosaicML'
_link = ''
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.curriculum_learning.curriculum_learning import CurriculumLearning as CurriculumLearning
from composer.algorithms.curriculum_learning.curriculum_learning import \
    CurriculumLearningHparams as CurriculumLearningHparams

_name = 'Curriculum Learning'
_class_name = 'CurriculumLearning'
_functional = 'apply_curriculum'
_tldr = 'Using sequence length as a proxy for example difficulty, it warms up the sequence length for a specified duration of training.'
_attribution = '(Li et al, 2021)'
_link = 'https://arxiv.org/abs/2108.06084'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import os
from unittest.mock import MagicMock, Mock

import pytest
import torch
import torch.distributed as dist
import torch.utils.data
from _pytest.monkeypatch import MonkeyPatch

from composer import Logger, State
from composer.core.types import DataLoader, Model, Precision, Tensors
from composer.datasets import DataloaderHparams, DataloaderSpec, DatasetHparams, SyntheticDatasetHparams
from composer.models import BaseMosaicModel, MnistClassifierHparams, ModelHparams, MosaicClassifier
from composer.optim import AdamHparams, ExponentialLRHparams
from composer.trainer import TrainerHparams
from composer.trainer.ddp import DDPDataLoader, DDPHparams, FileStoreHparams
from composer.trainer.devices import CPUDeviceHparams


@pytest.fixture()
def dummy_model_hparams() -> ModelHparams:
    return MnistClassifierHparams(num_classes=10)


@pytest.fixture()
def dummy_model(dummy_model_hparams: ModelHparams) -> BaseMosaicModel:
    return dummy_model_hparams.initialize_object()


@pytest.fixture()
def dummy_dataset_hparams() -> SyntheticDatasetHparams:
    return SyntheticDatasetHparams(num_classes=10,
                                   shape=[1, 28, 28],
                                   device="cpu",
                                   sample_pool_size=256,
                                   one_hot=False,
                                   drop_last=True,
                                   shuffle=False)


@pytest.fixture()
def dummy_dataloader_spec(dummy_dataset_hparams: SyntheticDatasetHparams) -> DataloaderSpec:
    return dummy_dataset_hparams.initialize_object()


@pytest.fixture()
def dummy_state_without_rank(dummy_model: BaseMosaicModel) -> State:
    state = State(
        model=dummy_model,
        epoch=5,
        step=50,
        precision=Precision.FP32,
        grad_accum=1,
        train_batch_size=10,
        eval_batch_size=10,
        max_epochs=10,
    )
    return state


@pytest.fixture
def dummy_dataloader_hparams() -> DataloaderHparams:
    return DataloaderHparams(
        num_workers=0,
        prefetch_factor=2,
        persistent_workers=False,
        pin_memory=False,
        timeout=0,
    )


def get_dataloader(dataloader_spec: DataloaderSpec, dataloader_hparams: DataloaderHparams) -> DataLoader:
    batch_size = 10

    sampler = torch.utils.data.DistributedSampler[int](
        dataloader_spec.dataset,
        drop_last=dataloader_spec.drop_last,
        shuffle=dataloader_spec.shuffle,
        rank=0,
        num_replicas=1,
    )

    dataloader = torch.utils.data.DataLoader(
        dataloader_spec.dataset,
        batch_size=batch_size,
        shuffle=False,  # set in the sampler
        num_workers=dataloader_hparams.num_workers,
        pin_memory=dataloader_hparams.pin_memory,
        drop_last=dataloader_spec.drop_last,
        sampler=sampler,
        collate_fn=dataloader_spec.collate_fn,
        worker_init_fn=dataloader_spec.worker_init_fn,
        multiprocessing_context=dataloader_spec.multiprocessing_context,
        generator=dataloader_spec.generator,
        timeout=dataloader_hparams.timeout,
        prefetch_factor=dataloader_hparams.prefetch_factor,
        persistent_workers=dataloader_hparams.persistent_workers,
    )
    return DDPDataLoader(dataloader)


@pytest.fixture
def dummy_train_dataloader(dummy_dataloader_spec: DataloaderSpec,
                           dummy_dataloader_hparams: DataloaderHparams) -> DataLoader:
    return get_dataloader(dummy_dataloader_spec, dummy_dataloader_hparams)


@pytest.fixture
def dummy_val_dataloader(dummy_dataloader_spec: DataloaderSpec,
                         dummy_dataloader_hparams: DataloaderHparams) -> DataLoader:
    return get_dataloader(dummy_dataloader_spec, dummy_dataloader_hparams)


@pytest.fixture()
def dummy_state(dummy_state_without_rank: State, monkeypatch: MonkeyPatch) -> State:
    monkeypatch.setattr(dist, "get_rank", lambda: 0)
    return dummy_state_without_rank


@pytest.fixture()
def dummy_state_dl(dummy_state: State, dummy_train_dataloader: DataLoader) -> State:
    dummy_state.train_dataloader = dummy_train_dataloader
    return dummy_state


@pytest.fixture()
def dummy_logger(dummy_state: State):
    return Logger(dummy_state)


@pytest.fixture
def logger_mock():
    return MagicMock()


"""
Dummy algorithms
"""


@pytest.fixture()
def algorithms(always_match_algorithms):
    return always_match_algorithms


@pytest.fixture()
def always_match_algorithms():
    attrs = {'match.return_value': True}
    return [Mock(**attrs) for _ in range(5)]


@pytest.fixture()
def never_match_algorithms():
    attrs = {'match.return_value': False}
    return [Mock(**attrs) for _ in range(5)]


@pytest.fixture
def mosaic_trainer_hparams(
    dummy_model_hparams: ModelHparams,
    dummy_dataset_hparams: DatasetHparams,
    ddp_tmpdir: str,
) -> TrainerHparams:
    return TrainerHparams(
        algorithms=[],
        optimizer=AdamHparams(),
        schedulers=[ExponentialLRHparams(gamma=0.1)],
        max_epochs=2,
        precision=Precision.FP32,
        total_batch_size=64,
        eval_batch_size=64,
        ddp=DDPHparams(
            store=FileStoreHparams(os.path.join(ddp_tmpdir, "store")),
            node_rank=0,
            num_nodes=1,
            fork_rank_0=False,
        ),
        dataloader=DataloaderHparams(
            num_workers=0,
            prefetch_factor=2,
            persistent_workers=False,
            pin_memory=False,
            timeout=0,
        ),
        device=CPUDeviceHparams(n_cpus=1),
        loggers=[],
        model=dummy_model_hparams,
        val_dataset=dummy_dataset_hparams,
        train_dataset=dummy_dataset_hparams,
        grad_accum=1,
    )


"""
Simple models
"""


class SimpleConvModel(torch.nn.Module):

    def __init__(self):
        super().__init__()

        conv_args = dict(kernel_size=(3, 3), padding=1)
        self.conv1 = torch.nn.Conv2d(in_channels=32, out_channels=8, stride=2, bias=False, **conv_args)  # stride > 1
        self.conv2 = torch.nn.Conv2d(in_channels=8, out_channels=32, stride=2, bias=False,
                                     **conv_args)  # stride > 1 but in_channels < 16
        self.conv3 = torch.nn.Conv2d(in_channels=32, out_channels=64, stride=1, bias=False, **conv_args)  # stride = 1

        self.pool1 = torch.nn.MaxPool2d(kernel_size=(2, 2), stride=2, padding=1)

    def forward(self, x: Tensors) -> Tensors:  # type: ignore
        # Very basic forward operation with no activation functions
        # used just to test that model surgery doesn't create forward prop bugs.
        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)
        out = self.pool1(out)
        return out


@pytest.fixture()
def simple_conv_model():
    return MosaicClassifier(SimpleConvModel())


@pytest.fixture()
def simple_conv_model_input():
    return torch.rand((64, 32, 64, 64))


@pytest.fixture()
def state_with_model(simple_conv_model: Model):
    state = State(
        epoch=50,
        step=50,
        train_batch_size=100,
        eval_batch_size=100,
        grad_accum=1,
        max_epochs=100,
        model=simple_conv_model,
        precision=Precision.FP32,
    )
    return state



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from typing import TYPE_CHECKING, Mapping

import transformers
from torchmetrics.collections import MetricCollection

from composer.models.nlp_metrics import Perplexity
from composer.models.transformer_shared import MosaicTransformer

if TYPE_CHECKING:
    from composer.core.types import Batch, Metrics, Tensors


class GPT2Model(MosaicTransformer):
    """
    Implements a GPT-2 wrapper around a MosaicTransformer.
    """

    def __init__(self, module: transformers.GPT2Model, config: transformers.GPT2Config, tokenizer_name: str) -> None:
        super().__init__(
            module=module,  #type: ignore (thirdparty)
            config=config,
            tokenizer_name=tokenizer_name)

        # If we ever have algorithms that modify the loss function, then this might be a bit inefficient
        #  because it'll compute the expensive softmax operation twice.
        # Instead, we should consider figuring out how to leverage self.train_loss and return the e^self.train_loss.
        # Of course, this also depends on the implementation details of algorithms.
        self.train_perplexity = Perplexity()
        self.val_perplexity = Perplexity()

    def loss(self, outputs: Mapping, batch: Batch) -> Tensors:

        if outputs.get('loss', None) is not None:
            return outputs['loss']
        else:
            raise NotImplementedError('Calculating loss directly not supported yet.')

    def metrics(self, train: bool = False) -> Metrics:
        return MetricCollection([self.train_loss, self.train_perplexity]) if train else MetricCollection(
            [self.val_loss, self.val_perplexity])



# Copyright 2021 MosaicML. All Rights Reserved.

from collections import Counter

import numpy as np
import pytest
import torch
from torch.optim.lr_scheduler import ExponentialLR, MultiStepLR

from composer.algorithms import ScaleSchedule
from composer.algorithms.scale_schedule import scale_scheduler
from composer.core.event import Event
from composer.core.state import State
from composer.core.types import Optimizer
from composer.loggers import Logger
from composer.optim.pytorch_future import WarmUpLR


@pytest.fixture
def optimizer(dummy_model):
    return torch.optim.SGD(dummy_model.parameters(), lr=1)


def flatten(lst: list):
    return [x for sublst in lst for x in sublst]


@pytest.mark.parametrize('ssr', [0.5, 0.75, 1.0])
class TestScaleSchedule():

    @staticmethod
    def _test(targets, scheduler, epochs, optimizer, ssr):
        scale_scheduler(scheduler, ssr)
        print(targets)
        for epoch in range(epochs):
            for param_group in optimizer.param_groups:
                print(f"target: {targets[epoch]}, actual: {param_group['lr']}")
                torch.testing.assert_allclose(targets[epoch], param_group['lr'])
            scheduler.step()

    def test_scale_schedule_step_lr(self, optimizer, ssr):
        epochs = int(9 * ssr)
        step_size = int(3 * ssr)
        gamma = 0.1
        targets = flatten([[1.0 * (gamma**n)] * step_size for n in range(30)])
        targets = targets[:epochs]

        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=3, gamma=0.1)
        self._test(targets, scheduler, epochs, optimizer, ssr)

    def test_scale_schedule_multistep_lr(self, optimizer, ssr):
        epochs = int(9 * ssr)
        milestones = np.diff([0, int(2 * ssr), int(7 * ssr), epochs])
        gamma = 0.1
        targets = flatten([[1.0 * (gamma**n)] * ms for n, ms in enumerate(milestones)])
        targets = targets[:epochs]

        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, [2, 7], gamma)
        self._test(targets, scheduler, epochs, optimizer, ssr)

    def test_scale_schedule_exponential(self, optimizer, ssr):
        epochs = int(9 * ssr)
        targets = [1.0 * 0.1**(x / ssr) for x in range(epochs)]
        scheduler = ExponentialLR(optimizer, gamma=0.1)
        self._test(targets, scheduler, epochs, optimizer, ssr)

    @pytest.mark.xfail
    def test_scale_schedule_cosine(self, optimizer: Optimizer, ssr: float):
        raise NotImplementedError

    @pytest.mark.xfail
    def test_scale_schedule_cosine_warm_restarts(self, optimizer: Optimizer, ssr: float):
        raise NotImplementedError

    def test_scale_schedule_warmup(self, optimizer: Optimizer, ssr: float):
        targets = [0.5] * 4 + [1.0] * 5  # no effect
        scheduler = WarmUpLR(optimizer, warmup_factor=0.5, warmup_iters=4, warmup_method='constant')
        epochs = int(9 * ssr)
        targets = targets[:epochs]
        self._test(targets, scheduler, epochs, optimizer, ssr)


@pytest.mark.parametrize('ssr', [0.5, 0.75, 1.0])
class TestScaleScheduleAlgorithm():

    def test_epochs_scaled(self, dummy_state: State, optimizer: Optimizer, ssr: float, noop_dummy_logger: Logger):

        scheduler = MultiStepLR(optimizer, milestones=[30, 50], gamma=0.1)
        dummy_state.schedulers = scheduler
        dummy_state.max_epochs = 10
        algorithm = ScaleSchedule(ratio=ssr)
        algorithm.apply(Event.TRAINING_START, dummy_state, noop_dummy_logger)
        assert dummy_state.max_epochs == int(10 * ssr)
        assert dummy_state.schedulers.milestones == Counter([int(30 * ssr), int(50 * ssr)])  # type: ignore

    @pytest.mark.xfail
    def test_scale_schedule_compose1(self, optimizer: Optimizer, ssr: float):
        return NotImplementedError

    @pytest.mark.xfail
    def test_scale_schedule_compose2(self, optimizer: Optimizer, ssr: float):
        return NotImplementedError


def test_epochs_validate_zero_epochs(dummy_state: State, noop_dummy_logger: Logger):
    algorithm = ScaleSchedule(ratio=0.01)
    dummy_state.max_epochs = 10
    dummy_state.schedulers = tuple()
    with pytest.raises(ValueError):
        algorithm.apply(Event.TRAINING_START, dummy_state, noop_dummy_logger)


def test_epochs_validate_run_once(dummy_state: State, noop_dummy_logger: Logger):
    algorithm = ScaleSchedule(ratio=0.1)
    dummy_state.schedulers = tuple()
    with pytest.raises(AssertionError):
        algorithm.apply(Event.TRAINING_START, dummy_state, noop_dummy_logger)
        algorithm.apply(Event.TRAINING_START, dummy_state, noop_dummy_logger)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.label_smoothing.label_smoothing import LabelSmoothing as LabelSmoothing
from composer.algorithms.label_smoothing.label_smoothing import LabelSmoothingHparams as LabelSmoothingHparams
from composer.algorithms.label_smoothing.label_smoothing import smooth_labels as smooth_labels

_name = 'Label Smoothing'
_class_name = 'LabelSmoothing'
_functional = 'smooth_labels'
_tldr = 'Smooths the labels with a uniform prior'
_attribution = '(Szegedy et al, 2015)'
_link = 'https://arxiv.org/abs/1512.00567'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest
import torch

from composer.algorithms.progressive_resizing import ProgressiveResizing, ProgressiveResizingHparams, resize_inputs
from composer.core import Event
from composer.trainer import TrainerHparams
from tests.utils.trainer_fit import train_model


def check_scaled_shape(orig: torch.Tensor, scaled: torch.Tensor, scale_factor: float) -> bool:
    """ Asserts that the scaled shape is correct, given orig shape and scale_factor"""
    N, C, H, W = orig.shape
    Hc = int(scale_factor * H)
    Wc = int(scale_factor * W)

    return scaled.shape == (N, C, Hc, Wc)


@pytest.fixture
def Wx() -> int:
    return 32


@pytest.fixture
def Hx(Wx) -> int:
    return Wx


@pytest.fixture
def X(Wx, Hx):
    torch.manual_seed(0)
    return torch.randn(16, 8, Hx, Wx)


@pytest.fixture
def Wy(Wx) -> int:
    return Wx


@pytest.fixture
def Hy(Hx) -> int:
    return Hx


@pytest.fixture
def y(Wy, Hy):
    torch.manual_seed(0)
    return torch.randn(16, 8, Hy, Wy)


@pytest.fixture(params=[0.5, 0.75, 1])
def scale_factor(request) -> float:
    return request.param


@pytest.fixture(params=["resize", "crop"])
def mode(request) -> str:
    return request.param


@pytest.fixture
def initial_scale() -> float:
    return 0.5


@pytest.fixture
def finetune_fraction() -> float:
    return 0.2


@pytest.fixture
def resize_targets() -> bool:
    return False


@pytest.fixture
def dummy_hparams(mode, initial_scale, finetune_fraction, resize_targets):
    return ProgressiveResizingHparams(mode, initial_scale, finetune_fraction, resize_targets)


@pytest.fixture
def dummy_algorithm(dummy_hparams):
    return ProgressiveResizing(dummy_hparams.mode, dummy_hparams.initial_scale, dummy_hparams.finetune_fraction,
                               dummy_hparams.resize_targets)


def test_resize_noop(X, y, mode):
    """ Tests that no operation is performed when scale_factor == 1"""
    Xc, yc = resize_inputs(X, y, 1.0, mode, resize_targets=False)
    assert X is Xc


@pytest.mark.parametrize("y", [None])
def test_without_target(X, y):
    """ Test that resizing works properly with no target present"""
    try:
        Xc, yc = resize_inputs(X, y, 1.0, "crop", resize_targets=False)
    except:
        pytest.fail("apply_progressive_resizing failed with y == None")


@pytest.mark.parametrize("Wx,Hx", [(31, 31), (32, 32), (32, 16)])
def test_resize_inputs_shape(X, y, mode, scale_factor):
    """ Test scaling works for different input shapes"""

    Xc, yc = resize_inputs(X, y, scale_factor, mode, resize_targets=False)
    assert check_scaled_shape(X, Xc, scale_factor)


def test_resize_outputs_shape(X, y, mode, scale_factor):
    """ Test that resizing outputs works """

    Xc, yc = resize_inputs(X, y, scale_factor, mode, resize_targets=True)
    assert check_scaled_shape(y, yc, scale_factor)


@pytest.mark.parametrize("Wx,Hx,Wy,Hy", [(32, 32, 16, 16)])
def test_resize_outputs_different_shape(X, y, scale_factor, mode):
    """ Test that resizing works when X and y have different shapes"""

    Xc, yc = resize_inputs(X, y, scale_factor, mode, resize_targets=True)
    assert check_scaled_shape(y, yc, scale_factor)


@pytest.mark.parametrize("mode,initial_scale,finetune_fraction", [("foo", 0.5, 0.2), ("crop", 1.2, 0.2),
                                                                  ("crop", 0.5, 1.2)])
def test_invalid_hparams(mode, initial_scale, finetune_fraction):
    """ Test that invalid hyperparameters error.
    Ideally this could be caught by the Hparams, but that's not yet supported in yahp.
    """
    with pytest.raises(ValueError):
        ProgressiveResizing(mode, initial_scale, finetune_fraction, False)


@pytest.mark.parametrize("event", [Event.AFTER_DATALOADER])
def test_match_correct(event, dummy_algorithm, dummy_state):
    """ Algo should match AFTER_DATALOADER """
    assert dummy_algorithm.match(event, dummy_state)


@pytest.mark.parametrize("event", [Event.TRAINING_START])
def test_match_incorrect(event, dummy_algorithm, dummy_state):
    """ Algo should NOT match TRAINING_START """
    assert not dummy_algorithm.match(event, dummy_state)


@pytest.mark.parametrize("epoch_frac", [0, 0.8, 1])
def test_apply(epoch_frac, X, y, dummy_algorithm, dummy_state, dummy_logger):
    """ Test apply at different epoch fractions (fraction of max epochs) """
    dummy_state.epoch = int(epoch_frac * dummy_state.max_epochs)
    s = dummy_algorithm.hparams.initial_scale
    f = dummy_algorithm.hparams.finetune_fraction
    scale_factor = min([s + (1 - s) / (1 - f) * epoch_frac, 1.0])
    dummy_state.batch = (X, y)
    dummy_algorithm.apply(Event.AFTER_DATALOADER, dummy_state, dummy_logger)

    last_input, _ = dummy_state.batch
    assert check_scaled_shape(X, last_input, scale_factor)


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_progressive_resizing_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [
        ProgressiveResizingHparams(mode="resize", initial_scale=0.5, finetune_fraction=0.2, resize_targets=False)
    ]
    train_model(mosaic_trainer_hparams)



# Copyright 2021 MosaicML. All Rights Reserved.

import os
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
import torch
import torch.utils.data
import yahp as hp
from PIL import Image
from torchvision import transforms
from torchvision.datasets import ImageFolder

from composer.core.types import Batch, Tensor
from composer.datasets.hparams import DataloaderSpec, DatasetHparams


class PreprocessingFn:

    def __init__(self) -> None:
        self.mean: Optional[Tensor] = None
        self.std: Optional[Tensor] = None

    def __call__(self, batch: Batch):
        xs, ys = batch
        assert isinstance(xs, Tensor)
        assert isinstance(ys, Tensor)
        device = xs.device

        if self.mean is None:
            self.mean = torch.tensor([0.485 * 255, 0.456 * 255, 0.406 * 255], device=device)
            self.mean = self.mean.view(1, 3, 1, 1)
        if self.std is None:
            self.std = torch.tensor([0.229 * 255, 0.224 * 255, 0.225 * 255], device=device)
            self.std = self.std.view(1, 3, 1, 1)

        xs = xs.float()
        xs = xs.sub_(self.mean).div_(self.std)
        return xs, ys


def fast_collate(batch: List[Tuple[Image.Image, Tensor]], memory_format: torch.memory_format = torch.contiguous_format):
    imgs = [img[0] for img in batch]
    targets = torch.tensor([target[1] for target in batch], dtype=torch.int64)
    w = imgs[0].size[0]
    h = imgs[0].size[1]
    tensor = torch.zeros((len(imgs), 3, h, w), dtype=torch.uint8).contiguous(memory_format=memory_format)
    for i, img in enumerate(imgs):
        nump_array = np.asarray(img, dtype=np.uint8)
        if nump_array.ndim < 3:
            nump_array = np.expand_dims(nump_array, axis=-1)

        nump_array = np.rollaxis(nump_array, 2).copy()
        if nump_array.shape[0] != 3:
            assert nump_array.shape[0] == 1, "unexpected shape"
            nump_array = np.resize(nump_array, (3, h, w))
        assert tuple(tensor.shape)[1:] == nump_array.shape, "shape mistmatch"

        tensor[i] += torch.from_numpy(nump_array)

    return tensor, targets


@dataclass
class ImagenetDatasetHparams(DatasetHparams):
    resize_size: int = hp.required("resize size")
    crop_size: int = hp.required("crop size")
    is_train: bool = hp.required("whether to load the training or validation dataset")
    datadir: str = hp.required("data directory")
    drop_last: bool = hp.optional("Whether to drop the last samples for the last batch", default=True)
    shuffle: bool = hp.optional("Whether to shuffle the dataset for each epoch", default=True)

    def initialize_object(self) -> DataloaderSpec:
        datadir = self.datadir
        is_train = self.is_train

        if is_train:
            # include fixed-size resize before RandomResizedCrop in training only
            # if requested (by specifying a size > 0)
            train_resize_size = self.resize_size
            train_transforms: List[torch.nn.Module] = []
            if train_resize_size > 0:
                train_transforms.append(transforms.Resize(train_resize_size))
            # always include RandomResizedCrop and RandomHorizontalFlip
            train_transforms += [
                transforms.RandomResizedCrop(self.crop_size, scale=(0.08, 1.0), ratio=(0.75, 4.0 / 3.0)),
                transforms.RandomHorizontalFlip()
            ]
            transformation = transforms.Compose(train_transforms)
        else:
            transformation = transforms.Compose([
                transforms.Resize(self.resize_size),
                transforms.CenterCrop(self.crop_size),
            ])

        split = "train" if is_train else "val"

        return DataloaderSpec(
            dataset=ImageFolder(os.path.join(datadir, split), transformation),
            drop_last=self.drop_last,
            collate_fn=fast_collate,
            shuffle=self.shuffle,
            prefetch_fn=PreprocessingFn(),
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.base import BaseMosaicModel as BaseMosaicModel
from composer.models.base import MosaicClassifier as MosaicClassifier
from composer.models.classify_mnist import MNIST_Classifier as MNIST_Classifier
from composer.models.classify_mnist import MnistClassifierHparams as MnistClassifierHparams
from composer.models.efficientnetb0 import EfficientNetB0 as EfficientNetB0
from composer.models.efficientnetb0 import EfficientNetB0Hparams as EfficientNetB0Hparams
from composer.models.gpt2 import GPT2Hparams as GPT2Hparams
from composer.models.model_hparams import Initializer as Initializer
from composer.models.model_hparams import ModelHparams as ModelHparams
from composer.models.resnet18 import ResNet18 as ResNet18
from composer.models.resnet18 import ResNet18Hparams as ResNet18Hparams
from composer.models.resnet50 import ResNet50 as ResNet50
from composer.models.resnet50 import ResNet50Hparams as ResNet50Hparams
from composer.models.resnet56_cifar10 import CIFAR10_ResNet56 as CIFAR10_ResNet56
from composer.models.resnet56_cifar10 import CIFARResNetHparams as CIFARResNetHparams
from composer.models.resnet101 import ResNet101 as ResNet101
from composer.models.resnet101 import ResNet101Hparams as ResNet101Hparams
from composer.models.transformer_shared import MosaicTransformer as MosaicTransformer
from composer.models.unet import UnetHparams as UnetHparams



# Copyright 2021 MosaicML. All Rights Reserved.

from torchvision import transforms


def add_dataset_transform(dataset, transform):
    """ Flexibly add the given transform to the dataset's collection of transforms

    Args:
        dataset: A torchvision-like dataset
        transform: Function to be added to the dataset's collection of transforms

    Returns:
        The original dataset. The transform is added in-place.
    """

    if not hasattr(dataset, "transform"):
        raise ValueError(f"Dataset of type {type(dataset)} has no attribute 'transform'. Expected TorchVision dataset.")

    if dataset.transform is None:
        dataset.transform = transform
    elif hasattr(dataset.transform, "transforms"):  # transform is a Compose
        dataset.transform.transforms.append(transform)
    else:  # transform is some other basic transform, join using Compose
        dataset.transform = transforms.Compose([dataset.transform, transform])

    return dataset



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.ghost_batchnorm.ghost_batchnorm import GhostBatchNorm as GhostBatchNorm
from composer.algorithms.ghost_batchnorm.ghost_batchnorm import GhostBatchNormHparams as GhostBatchNormHparams
from composer.algorithms.ghost_batchnorm.ghost_batchnorm import apply_ghost_batchnorm as apply_ghost_batchnorm



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.optim.decoupled_weight_decay import DecoupledAdamW as DecoupledAdamW
from composer.optim.decoupled_weight_decay import DecoupledSGDW as DecoupledSGDW
from composer.optim.optimizer_hparams import AdamHparams as AdamHparams
from composer.optim.optimizer_hparams import AdamWHparams as AdamWHparams
from composer.optim.optimizer_hparams import DecoupledAdamWHparams as DecoupledAdamWHparams
from composer.optim.optimizer_hparams import DecoupledSGDWHparams as DecoupledSGDWHparams
from composer.optim.optimizer_hparams import OptimizerHparams as OptimizerHparams
from composer.optim.optimizer_hparams import RAdamHparams as RAdamHparams
from composer.optim.optimizer_hparams import RMSPropHparams as RMSPropHparams
from composer.optim.optimizer_hparams import SGDHparams as SGDHparams
from composer.optim.pytorch_future import WarmUpLR as WarmUpLR
from composer.optim.scheduler import ComposedScheduler as ComposedScheduler
from composer.optim.scheduler import ConstantLR as ConstantLR
from composer.optim.scheduler import ConstantLRHparams as ConstantLRHparams
from composer.optim.scheduler import CosineAnnealingLRHparams as CosineAnnealingLRHparams
from composer.optim.scheduler import CosineAnnealingWarmRestartsHparams as CosineAnnealingWarmRestartsHparams
from composer.optim.scheduler import ExponentialLRHparams as ExponentialLRHparams
from composer.optim.scheduler import MultiStepLRHparams as MultiStepLRHparams
from composer.optim.scheduler import SchedulerHparams as SchedulerHparams
from composer.optim.scheduler import StepLRHparams as StepLRHparams
from composer.optim.scheduler import WarmUpLRHparams as WarmUpLRHparams



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import abc
from functools import wraps
from types import MethodType
from typing import TYPE_CHECKING, Any, Callable

from composer.core.serializable import Serializable
from composer.utils.ddp import is_rank_set, is_rank_zero

if TYPE_CHECKING:
    from composer import Logger, State


class Callback(Serializable, abc.ABC):
    """Base class for callbacks.
    
    Callbacks are similar to Algorithms, in that
    they are run on specific events. By convention, Callbacks should not
    modify ``State``.

    Each method name in ``Callback`` corresponds to an ``Event`` value.
    """

    def init(self, state: State, logger: Logger) -> None:
        pass

    def training_start(self, state: State, logger: Logger) -> None:
        pass

    def epoch_start(self, state: State, logger: Logger) -> None:
        pass

    def batch_start(self, state: State, logger: Logger) -> None:
        pass

    def after_dataloader(self, state: State, logger: Logger) -> None:
        pass

    def before_train_batch(self, state: State, logger: Logger) -> None:
        pass

    def before_forward(self, state: State, logger: Logger) -> None:
        pass

    def after_forward(self, state: State, logger: Logger) -> None:
        pass

    def before_loss(self, state: State, logger: Logger) -> None:
        pass

    def after_loss(self, state: State, logger: Logger) -> None:
        pass

    def before_backward(self, state: State, logger: Logger) -> None:
        pass

    def after_backward(self, state: State, logger: Logger) -> None:
        pass

    def after_train_batch(self, state: State, logger: Logger) -> None:
        pass

    def batch_end(self, state: State, logger: Logger) -> None:
        pass

    def epoch_end(self, state: State, logger: Logger) -> None:
        pass

    def training_end(self, state: State, logger: Logger) -> None:
        pass

    def eval_start(self, state: State, logger: Logger) -> None:
        pass

    def eval_batch_start(self, state: State, logger: Logger) -> None:
        pass

    def eval_before_forward(self, state: State, logger: Logger) -> None:
        pass

    def eval_after_forward(self, state: State, logger: Logger) -> None:
        pass

    def eval_batch_end(self, state: State, logger: Logger) -> None:
        pass

    def eval_end(self, state: State, logger: Logger) -> None:
        pass


class RankZeroCallback(Callback, abc.ABC):
    """
    Base class for callbacks that should only run on the rank 0 process.

    This ensures that after a DDP fork, callback methods will execute only once per rank. Note that :meth:`init`
    and :meth:`load_state_dict` are executed before the DDP fork and will be called on all ranks.
    """

    def __init__(self) -> None:
        from composer.core import Event

        super().__init__()

        # ensure all callbacks are executed only on rank 0
        functions_to_wrap = [*(event.value for event in Event), "state_dict"]

        for fn_name in functions_to_wrap:
            original_fn = getattr(self, fn_name)

            @wraps(original_fn)
            def wrapped_fn(
                backend: RankZeroCallback,
                *args: Any,
                original_fn: Callable[[State, Logger], None] = original_fn,
                **kwargs: Any,
            ) -> None:
                if is_rank_set():
                    if not is_rank_zero():
                        return
                return original_fn(*args, **kwargs)

            setattr(self, fn_name, MethodType(wrapped_fn, self))
